"""_xml.py: functions to support xml creation methods in the Model class."""

# following should be kept in line with major.minor tag values in repository
citation_format = 'bp:resqpy:1.6'
use_version_string = False

import logging

log = logging.getLogger(__name__)

import getpass
import os
import warnings

import resqpy.crs as rqc
import resqpy.olio.time as time
import resqpy.olio.uuid as bu
import resqpy.olio.xml_et as rqet
from resqpy.olio.xml_namespaces import curly_namespace as ns
from resqpy.olio.xml_namespaces import namespace as ns_url


def _create_tree_if_none(model):
    """Checks that model has an xml tree; if not, an empty tree is created; not usually called directly."""

    if model.main_tree is None:
        model.main_tree = rqet.ElementTree()
        model.modified = True


def _change_uuid_in_supporting_representation_reference(model, node, old_uuid, new_uuid, new_title = None):
    """Look for supporting representation reference using the old_uuid and replace with the new_uuid."""

    ref_node = rqet.find_tag(node, 'SupportingRepresentation')
    if ref_node is None:
        return False
    uuid_node = rqet.find_tag(ref_node, 'UUID')
    if uuid_node is None:
        return False
    if not bu.matching_uuids(uuid_node.text, old_uuid):
        return False
    uuid_node.text = str(new_uuid)
    if new_title:
        title_node = rqet.find_tag(ref_node, 'Title')
        if title_node is not None:
            title_node.text = str(new_title)
    model.set_modified()
    return True


def _create_root(model):
    """Initialises an empty main xml tree for model."""

    assert (model.main_tree is None)
    assert (model.main_root is None)
    model.main_root = rqet.Element(ns['content_types'] + 'Types')
    model.main_tree = rqet.ElementTree(element = model.main_root)


def _new_obj_node(flavour, name_space = 'resqml2', is_top_lvl_obj = True):
    """Creates a new main object element and sets attributes (does not add children)."""

    if flavour.startswith('obj_'):
        flavour = flavour[4:]

    node = rqet.Element(ns[name_space] + flavour)
    node.set('schemaVersion', '2.0')
    node.set('uuid', str(bu.new_uuid()))
    if is_top_lvl_obj:
        node.set(ns['xsi'] + 'type', ns[name_space] + 'obj_' + flavour)
    node.text = rqet.null_xml_text

    return node


def _referenced_node(model, ref_node, consolidate = False):
    """For a given xml reference node, returns the node for the object referred to, if present."""

    if ref_node is None:
        return None
    #      content_type = rqet.find_tag_text(ref_node, 'ContentType')
    uuid = bu.uuid_from_string(rqet.find_tag_text(ref_node, 'UUID'))
    if uuid is None:
        return None
    #      return model.root_for_part(model.parts_list_of_type(type_of_interest = content_type, uuid = uuid))
    if consolidate and model.consolidation is not None and uuid in model.consolidation.map:
        resident_uuid = model.consolidation.map[uuid]
        if resident_uuid is None:
            return None
        node = model.root_for_part(model.part_for_uuid(resident_uuid))
        if node is not None:
            # patch resident uuid and title into ref node!
            uuid_node = rqet.find_tag(ref_node, 'UUID')
            uuid_node.text = str(resident_uuid)
            title_node = rqet.find_tag(ref_node, 'Title')
            if title_node is not None:
                title = rqet.citation_title_for_node(node)
                if title:
                    title_node.text = str(title)
    else:
        node = model.root_for_part(model.part_for_uuid(uuid))
    return node


def _create_ref_node(flavour, title, uuid, content_type = None, root = None):
    """Create a reference node, optionally add to root."""

    assert uuid is not None

    if flavour.startswith('obj_'):
        flavour = flavour[4:]

    if not content_type:
        content_type = 'obj_' + flavour
    else:
        if content_type[0].isupper():
            content_type = 'obj_' + content_type

    prefix = ns['eml'] if flavour == 'HdfProxy' else ns['resqml2']
    ref_node = rqet.Element(prefix + flavour)
    ref_node.set(ns['xsi'] + 'type', ns['eml'] + 'DataObjectReference')
    ref_node.text = rqet.null_xml_text

    ct_node = rqet.SubElement(ref_node, ns['eml'] + 'ContentType')
    ct_node.set(ns['xsi'] + 'type', ns['xsd'] + 'string')
    if 'EpcExternalPartReference' in content_type:
        ct_node.text = 'application/x-eml+xml;version=2.0;type=' + content_type
    else:
        ct_node.text = 'application/x-resqml+xml;version=2.0;type=' + content_type

    if not title:
        title = '(title unavailable)'
    title_node = rqet.SubElement(ref_node, ns['eml'] + 'Title')
    title_node.set(ns['xsi'] + 'type', ns['eml'] + 'DescriptionString')
    title_node.text = title

    uuid_node = rqet.SubElement(ref_node, ns['eml'] + 'UUID')
    uuid_node.set(ns['xsi'] + 'type', ns['eml'] + 'UuidString')
    uuid_node.text = str(uuid)

    if use_version_string:
        version_str = rqet.SubElement(ref_node, ns['eml'] + 'VersionString')  # I'm guessing what this is
        version_str.set(ns['xsi'] + 'type', ns['eml'] + 'NameString')
        version_str.text = bu.version_string(uuid)

    if root is not None:
        root.append(ref_node)

    return ref_node


def _uom_node(root, uom):
    """Add a generic unit of measure sub element to root."""

    assert root is not None and uom is not None and len(uom)
    # todo: could assert that uom is a valid unit of measure

    node = rqet.SubElement(root, ns['resqml2'] + 'UOM')
    node.set(ns['xsi'] + 'type', ns['resqml2'] + 'ResqmlUom')
    node.text = uom

    return node


def _create_rels_part(model):
    """Adds a relationships reference node as a new part in the model's parts forest."""

    rels = rqet.SubElement(model.main_root, ns['content_types'] + 'Default')
    rels.set('Extension', 'rels')
    rels.set('ContentType', 'application/vnd.openxmlformats-package.relationships+xml')
    model.rels_present = True
    model.set_modified()

    return rels


def _create_citation(root = None, title = '', originator = None):
    """Creates a citation xml node and optionally appends as a child of root."""

    if not title:
        title = '(no title)'

    citation = rqet.Element(ns['eml'] + 'Citation')
    citation.set(ns['xsi'] + 'type', ns['eml'] + 'Citation')
    citation.text = rqet.null_xml_text

    title_node = rqet.SubElement(citation, ns['eml'] + 'Title')
    title_node.set(ns['xsi'] + 'type', ns['eml'] + 'DescriptionString')
    title_node.text = title

    originator_node = rqet.SubElement(citation, ns['eml'] + 'Originator')
    if originator is None:
        try:
            originator = str(getpass.getuser())
        except Exception:
            originator = 'unknown'
    originator_node.set(ns['xsi'] + 'type', ns['eml'] + 'NameString')
    originator_node.text = originator

    creation_node = rqet.SubElement(citation, ns['eml'] + 'Creation')
    creation_node.set(ns['xsi'] + 'type', ns['xsd'] + 'dateTime')
    creation_node.text = time.now()

    format_node = rqet.SubElement(citation, ns['eml'] + 'Format')
    format_node.set(ns['xsi'] + 'type', ns['eml'] + 'DescriptionString')
    if rqet.pretend_to_be_fesapi:
        format_node.text = '[F2I-CONSULTING:fesapi]'
    else:
        format_node.text = citation_format

    # todo: add optional description field

    if root is not None:
        root.append(citation)

    return citation


def _create_unknown(root = None):
    """Creates an Unknown node and optionally adds as child of root."""

    unknown = rqet.Element(ns['eml'] + 'Unknown')
    unknown.set(ns['xsi'] + 'type', ns['eml'] + 'DescriptionString')
    unknown.text = 'Unknown'
    if root is not None:
        root.append(unknown)
    return unknown


def _create_doc_props(model, add_as_part = True, root = None, originator = None):
    """Creates a document properties stub node and optionally adds as child of root and/or to parts forest."""

    dp = rqet.Element(ns['cp'] + 'coreProperties')
    dp.text = rqet.null_xml_text

    created = rqet.SubElement(dp, ns['dcterms'] + 'created')
    created.set(ns['xsi'] + 'type', ns['dcterms'] + 'W3CDTF')  # not sure of namespace here
    created.text = time.now()

    if originator is None:
        try:
            originator = str(os.getlogin())
        except Exception:
            originator = 'unknown'
    creator = rqet.SubElement(dp, ns['dc'] + 'creator')
    creator.text = originator

    ver = rqet.SubElement(dp, ns['cp'] + 'version')
    ver.text = '1.0'

    if root is not None:
        root.append(dp)
    if add_as_part:
        model.add_part('docProps', None, dp)
        if model.rels_present:
            (_, rel_tree) = model.rels_forest['_rels/.rels']
            core_rel = rqet.SubElement(rel_tree.getroot(), ns['rels'] + 'Relationship')
            core_rel.set('Id', 'CoreProperties')
            core_rel.set('Type', ns_url['rels_md'] + 'core-properties')
            core_rel.set('Target', 'docProps/core.xml')
    return dp


def _create_crs(model,
                add_as_part = True,
                title = 'cell grid local CRS',
                epsg_code = None,
                originator = None,
                x_offset = 0.0,
                y_offset = 0.0,
                z_offset = 0.0,
                areal_rotation_radians = 0.0,
                xy_units = 'm',
                z_units = 'm',
                z_inc_down = True):
    """DEPRECATED: Creates a Coordinate Reference System node and optionally adds as child of root and/or to parts forest."""

    crs = rqc.Crs(model,
                  x_offset = x_offset,
                  y_offset = y_offset,
                  z_offset = z_offset,
                  rotation = areal_rotation_radians,
                  xy_units = xy_units,
                  z_units = z_units,
                  z_inc_down = z_inc_down,
                  epsg_code = epsg_code)

    crs_node = crs.create_xml(add_as_part = add_as_part, title = title, originator = originator)

    if model.crs_uuid is None:
        model.crs_uuid = crs.uuid

    return crs_node


def _create_crs_reference(model, crs_root = None, root = None, crs_uuid = None):
    """Creates a node refering to an existing crs node and optionally adds as child of root."""

    if crs_uuid is None:
        warnings.warn('use of crs_root is deprecated in Model.create_crs_reference(); use crs_uuid instead')
        crs_uuid = rqet.uuid_for_part_root(crs_root)
    else:
        crs_root = model.root_for_uuid(crs_uuid)
    assert crs_root is not None

    return _create_ref_node('LocalCrs',
                            rqet.find_nested_tags_text(crs_root, ['Citation', 'Title']),
                            crs_uuid,
                            content_type = 'obj_LocalDepth3dCrs',
                            root = root)


def _create_md_datum_reference(md_datum_root, root = None):
    """Creates a node refering to an existing measured depth datum and optionally adds as child of root."""

    return _create_ref_node('MdDatum',
                            rqet.find_nested_tags_text(md_datum_root, ['Citation', 'Title']),
                            bu.uuid_from_string(md_datum_root.attrib['uuid']),
                            content_type = 'obj_MdDatum',
                            root = root)


def _create_hdf5_ext(model,
                     add_as_part = True,
                     root = None,
                     title = 'Hdf Proxy',
                     originator = None,
                     file_name = None,
                     uuid = None):
    """Creates an hdf5 external node and optionally adds as child of root and/or to parts forest."""

    ext = _new_obj_node('EpcExternalPartReference', name_space = 'eml')
    assert ext is not None
    if uuid is not None:  # preserve ext uuid if supplied
        ext.set('uuid', str(uuid))

    _create_citation(root = ext, title = title, originator = originator)

    mime_type = rqet.SubElement(ext, ns['eml'] + 'MimeType')
    mime_type.set(ns['xsi'] + 'type', ns['xsd'] + 'string')
    mime_type.text = 'application/x-hdf5'

    if root is not None:
        root.append(ext)
    if add_as_part:
        ext_uuid = bu.uuid_from_string(ext.attrib['uuid'])
        model.add_part('obj_EpcExternalPartReference', ext_uuid, ext)
        if not file_name:
            file_name = model.h5_file_name(override = 'full', file_must_exist = False)
        elif os.sep not in file_name:
            file_name = os.path.join(model.epc_directory, file_name)
        assert file_name
        log.debug(f'creating ext part for hdf5 file: {file_name}')
        model.h5_dict[ext_uuid.bytes] = file_name
        if model.main_h5_uuid is None:
            model.main_h5_uuid = ext_uuid
        if model.rels_present and file_name:
            (uuid, rel_tree) = model.rels_forest[rqet.rels_part_name_for_part(
                rqet.part_name_for_object('obj_EpcExternalPartReference', ext_uuid))]
            assert (bu.matching_uuids(uuid, ext_uuid))
            rel_node = rqet.SubElement(rel_tree.getroot(), ns['rels'] + 'Relationship')
            rel_node.set('Id', 'Hdf5File')
            rel_node.set('Type', ns_url['rels_ext'] + 'externalResource')
            rel_node.set('Target', file_name)
            rel_node.set('TargetMode', 'External')
    return ext


def _create_hdf5_dataset_ref(hdf5_uuid, object_uuid, group_tail, root, title = 'Hdf Proxy'):
    """Creates a pair of nodes referencing an hdf5 dataset (array) and adds to root."""

    assert root is not None
    assert group_tail

    if group_tail[0] == '/':
        group_tail = group_tail[1:]
    if group_tail[-1] == '/':
        group_tail = group_tail[:-1]
    hdf5_path = '/RESQML/' + str(object_uuid) + '/' + group_tail

    path_node = rqet.Element(ns['eml'] + 'PathInHdfFile')
    path_node.set(ns['xsi'] + 'type', ns['xsd'] + 'string')
    path_node.text = hdf5_path
    root.append(path_node)

    _create_ref_node('HdfProxy', title, hdf5_uuid, content_type = 'obj_EpcExternalPartReference', root = root)

    return path_node


def _create_supporting_representation(model,
                                      support_root = None,
                                      support_uuid = None,
                                      root = None,
                                      title = None,
                                      content_type = 'obj_IjkGridRepresentation'):
    """Craate a supporting representation reference node refering to an IjkGrid and optionally add to root."""

    assert support_root is not None or support_uuid is not None

    # todo: check that support_root is for a RESQML class that can support properties, matching content_type

    if support_root is not None:
        uuid = rqet.uuid_for_part_root(support_root)
        if uuid is not None:
            support_uuid = uuid
        if title is None:
            title = rqet.citation_title_for_node(support_root)
    assert support_uuid is not None
    if not title:
        title = model.title(uuid = support_uuid)
        if not title:
            title = 'supporting representation'

    return _create_ref_node('SupportingRepresentation', title, support_uuid, content_type = content_type, root = root)


def _create_source(source, root = None):
    """Create an extra meta data node holding information on the source of the data, optionally add to root."""

    emd_node = rqet.Element(ns['resqml2'] + 'ExtraMetadata')
    emd_node.set(ns['xsi'] + 'type', ns['resqml2'] + 'NameValuePair')
    emd_node.text = rqet.null_xml_text

    name_node = rqet.SubElement(emd_node, ns['resqml2'] + 'Name')
    name_node.set(ns['xsi'] + 'type', ns['xsd'] + 'string')
    name_node.text = 'source'

    value_node = rqet.SubElement(emd_node, ns['resqml2'] + 'Value')
    value_node.set(ns['xsi'] + 'type', ns['xsd'] + 'string')
    value_node.text = source

    if root is not None:
        root.append(emd_node)

    return emd_node


def _create_patch(p_uuid,
                  ext_uuid = None,
                  root = None,
                  patch_index = 0,
                  hdf5_type = 'DoubleHdf5Array',
                  xsd_type = 'double',
                  null_value = None,
                  const_value = None,
                  const_count = None,
                  points = False):
    """Create a node for a patch of values, including ref to hdf5 data set, optionally add to root."""

    if const_value is None:
        assert ext_uuid is not None
    else:
        assert const_count is not None and const_count > 0
        if hdf5_type.endswith('Hdf5Array'):
            hdf5_type = hdf5_type[:-9] + 'ConstantArray'

    lxt = str(xsd_type).lower()
    discrete = ('int' in lxt) or ('bool' in lxt)

    if points:
        assert not discrete
        patch_node = rqet.Element(ns['resqml2'] + 'PatchOfPoints')
        patch_node.set(ns['xsi'] + 'type', ns['resqml2'] + 'PatchOfPoints')
        patch_node.text = rqet.null_xml_text
        outer_values_tag = 'Points'
        inner_values_tag = 'Coordinates'
        hdf_path_tail = 'points_patch'
    else:
        patch_node = rqet.Element(ns['resqml2'] + 'PatchOfValues')
        patch_node.set(ns['xsi'] + 'type', ns['resqml2'] + 'PatchOfValues')
        patch_node.text = rqet.null_xml_text
        outer_values_tag = 'Values'
        inner_values_tag = 'Values'
        hdf_path_tail = 'values_patch'

    rep_patch_index = rqet.SubElement(patch_node, ns['resqml2'] + 'RepresentationPatchIndex')
    rep_patch_index.set(ns['xsi'] + 'type', ns['xsd'] + 'nonNegativeInteger')
    rep_patch_index.text = str(patch_index)

    outer_values_node = rqet.SubElement(patch_node, ns['resqml2'] + outer_values_tag)
    outer_values_node.set(ns['xsi'] + 'type', ns['resqml2'] + hdf5_type)  # may also be constant array type
    outer_values_node.text = rqet.null_xml_text

    if discrete:
        if null_value is None:
            if str(xsd_type).startswith('u'):
                null_value = 4294967295  # 2^32 - 1, used as default even for 64 bit data!
            else:
                null_value = -1
        null_value_node = rqet.SubElement(outer_values_node, ns['resqml2'] + 'NullValue')
        null_value_node.set(ns['xsi'] + 'type', ns['xsd'] + xsd_type)
        null_value_node.text = str(null_value)

    if const_value is None:

        inner_values_node = rqet.SubElement(outer_values_node, ns['resqml2'] + inner_values_tag)
        inner_values_node.set(ns['xsi'] + 'type', ns['eml'] + 'Hdf5Dataset')
        inner_values_node.text = rqet.null_xml_text

        _create_hdf5_dataset_ref(ext_uuid, p_uuid, f'{hdf_path_tail}{patch_index}', root = inner_values_node)

    else:

        const_value_node = rqet.SubElement(outer_values_node, ns['resqml2'] + 'Value')
        const_value_node.set(ns['xsi'] + 'type', ns['xsd'] + xsd_type)
        const_value_node.text = str(const_value)

        const_count_node = rqet.SubElement(outer_values_node, ns['resqml2'] + 'Count')
        const_count_node.set(ns['xsi'] + 'type', ns['xsd'] + 'nonNegativeInteger')
        const_count_node.text = str(const_count)

    if root is not None:
        root.append(patch_node)

    return patch_node


def _create_solitary_point3d(flavour, root, xyz):
    """Creates a subelement to root for a solitary point in 3D space."""

    # todo: check namespaces
    p3d = rqet.SubElement(root, ns['resqml2'] + flavour)
    p3d.set(ns['xsi'] + 'type', ns['resqml2'] + 'Point3d')
    p3d.text = rqet.null_xml_text

    for axis in range(3):
        coord_node = rqet.SubElement(p3d, ns['resqml2'] + 'Coordinate' + str(axis + 1))
        coord_node.set(ns['xsi'] + 'type', ns['xsd'] + 'double')
        coord_node.text = str(xyz[axis])

    return p3d


def _create_reciprocal_relationship(model, node_a, rel_type_a, node_b, rel_type_b, avoid_duplicates = True):
    """Adds a node to each of a pair of trees in the rels forest, to represent a two-way relationship."""

    def id_str(uuid):
        stringy = str(uuid)
        if not (rqet.pretend_to_be_fesapi or rqet.use_fesapi_quirks) or not stringy[0].isdigit():
            return stringy
        return '_' + stringy

    assert (model.rels_present)

    if node_a is None or node_b is None:
        log.error('attempt to create relationship with missing object')
        return

    uuid_a = node_a.attrib['uuid']
    obj_type_a = rqet.stripped_of_prefix(rqet.content_type(node_a.attrib[ns['xsi'] + 'type']))
    part_name_a = rqet.part_name_for_object(obj_type_a, uuid_a)
    rel_part_name_a = rqet.rels_part_name_for_part(part_name_a)
    (rel_uuid_a, rel_tree_a) = model.rels_forest[rel_part_name_a]
    rel_root_a = rel_tree_a.getroot()

    uuid_b = node_b.attrib['uuid']
    obj_type_b = rqet.stripped_of_prefix(rqet.content_type(node_b.attrib[ns['xsi'] + 'type']))
    part_name_b = rqet.part_name_for_object(obj_type_b, uuid_b)
    rel_part_name_b = rqet.rels_part_name_for_part(part_name_b)
    (rel_uuid_b, rel_tree_b) = model.rels_forest[rel_part_name_b]
    rel_root_b = rel_tree_b.getroot()

    create_a = True
    if avoid_duplicates:
        existing_rel_nodes = rqet.list_of_tag(rel_root_a, 'Relationship')
        for existing in existing_rel_nodes:
            if (rqet.stripped_of_prefix(existing.attrib['Type']) == rel_type_a and
                    existing.attrib['Target'] == part_name_b):
                create_a = False
                break
    if create_a:
        rel_a = rqet.SubElement(rel_root_a, ns['rels'] + 'Relationship')
        rel_a.set('Id',
                  id_str(uuid_b))  # NB: fesapi prefixes uuid with _ for some rels only (where uuid starts with a digit)
        rel_a.set('Type', ns_url['rels_ext'] + rel_type_a)
        rel_a.set('Target', part_name_b)

    create_b = True
    if avoid_duplicates:
        existing_rel_nodes = rqet.list_of_tag(rel_root_b, 'Relationship')
        for existing in existing_rel_nodes:
            if (rqet.stripped_of_prefix(existing.attrib['Type']) == rel_type_b and
                    existing.attrib['Target'] == part_name_a):
                create_b = False
                break
    if create_b:
        rel_b = rqet.SubElement(rel_root_b, ns['rels'] + 'Relationship')
        rel_b.set('Id',
                  id_str(uuid_a))  # NB: fesapi prefixes uuid with _ for some rels only (where uuid starts with a digit)
        rel_b.set('Type', ns_url['rels_ext'] + rel_type_b)
        rel_b.set('Target', part_name_a)



import math as maths

import numpy as np
import pytest
from numpy.testing import assert_array_almost_equal

import resqpy.crs as rqc
import resqpy.lines as rql
import resqpy.model as rq
import resqpy.olio.triangulation as tri
import resqpy.olio.vector_utilities as vec
import resqpy.surface as rqs
from resqpy.olio.random_seed import seed


def test_ccc():

    def check_equidistant(c, p_list):
        v = np.stack(tuple([c - p[:2] for p in p_list]))
        r = vec.naive_2d_lengths(v)
        assert_array_almost_equal(r[1:], r[0])

    # 3 points in orthogonal pattern
    p1 = np.array((0.0, 0.0, 0.0))
    p2 = np.array((20.0, 0.0, 0.0))
    p3 = np.array((0.0, 10.0, 5.0))
    assert_array_almost_equal(tri.ccc(p1, p2, p3), (10.0, 5.0))
    # equilateral triangle
    s = 23.57
    p1 = np.array((10.0, 20.0, 30.0))
    p2 = np.array((10.0, 20.0 + s, -45.0))
    p3 = np.array((10.0 + s * maths.cos(maths.radians(30.0)), 20.0 + 0.5 * s, 12.13))
    assert_array_almost_equal(tri.ccc(p1, p2, p3), np.mean(np.stack((p1, p2, p3)), axis = 0)[:2])
    # asymmetric triangle
    p1 = np.array((25.3, 12.1, 0.0))
    p2 = np.array((23.6, 2.9, -1.0))
    p3 = np.array((22.1, 87.3, 1.5))
    c = np.array(tri.ccc(p1, p2, p3))
    check_equidistant(c, (p1, p2, p3))
    # highly asymmetric triangle
    p1 = np.array((0.0, 0.0, 0.0))
    p2 = np.array((100.0, 0.0, 0.0))
    p3 = np.array((200.0, 1.0, 0.0))
    c = np.array(tri.ccc(p1, p2, p3))
    check_equidistant(c, (p1, p2, p3))


def test_voronoi():
    seed_value = 3567
    n_list = range(5, 50, 3)
    model = rq.Model(create_basics = True)
    crs = rqc.Crs(model)
    crs.create_xml()
    # setup unit square area of interest
    aoi_xyz = np.zeros((4, 3))
    aoi_xyz[1, 1] = 1.0
    aoi_xyz[2, :2] = 1.0
    aoi_xyz[3, 0] = 1.0
    aoi = rql.Polyline(model, set_coord = aoi_xyz, set_bool = True, set_crs = crs.uuid, title = 'aoi')
    # and an alternative area of interest
    aoi_heptagon_xyz = np.zeros((7, 3))
    aoi_heptagon_xyz[:, :2] = ((-0.5, -1.0), (-1.5, 0.5), (-1.0, 1.7), (0.5, 2.0), (2.0, 1.7), (2.5, 0.5), (1.5, -1.0))
    aoi_heptagon = rql.Polyline(model,
                                set_coord = aoi_heptagon_xyz,
                                set_bool = True,
                                set_crs = crs.uuid,
                                title = 'heptagon')
    aoi_heptagon_area = aoi_heptagon.area()

    for n in n_list:  # number of seed points (number of Voronoi cells)
        seed(seed_value)
        x = np.random.random(n)
        y = np.random.random(n)
        p = np.stack((x, y), axis = -1)
        # compute the Delauney triangulation
        t, b = tri.dt(p, plot_fn = None, progress_fn = None, return_hull = True)
        # dt function can return triangulation with a slightly concave hull, which voronoi function cannot handle
        hull = rql.Polyline(model, set_coord = p[b], set_bool = True, set_crs = crs.uuid, title = 'v cell')
        if not hull.is_convex():
            continue
        # test the Voronoi diagram with the unit square area of interest
        c, v = tri.voronoi(p, t, b, aoi)
        assert len(v) == n
        # check that the areas of the Voronoi cells sum to the area of interest
        area = 0.0
        for nodes in v:
            v_cell = rql.Polyline(model, set_coord = c[nodes], set_bool = True, set_crs = crs.uuid, title = 'v cell')
            area += v_cell.area()
        assert maths.isclose(area, 1.0, rel_tol = 0.001)
        # test the Voronoi diagram with the heptagonal area of interest
        c_hept, v_hept = tri.voronoi(p, t, b, aoi_heptagon)
        assert len(v_hept) == n
        area = 0.0
        for nodes in v_hept:
            v_cell = rql.Polyline(model,
                                  set_coord = c_hept[nodes],
                                  set_bool = True,
                                  set_crs = crs.uuid,
                                  title = 'v cell')
            area += v_cell.area()
        assert maths.isclose(area, aoi_heptagon_area, rel_tol = 0.001)
        # test re-triangulation of a Voronoi diagram, passing centre points
        points, triangles = tri.triangulated_polygons(c_hept, v_hept, centres = p)
        assert len(points) == len(c_hept) + n
        assert len(triangles) == sum(len(vh) for vh in v_hept)
        # test re-triangulation of a Voronoi diagram, computing centre points
        points_2, triangles_2 = tri.triangulated_polygons(c_hept, v_hept)
        assert len(points_2) == len(c_hept) + n
        assert len(triangles_2) == sum(len(vh) for vh in v_hept)
        assert np.all(triangles_2 == triangles)
        for cell in range(n):
            v_cell = rql.Polyline(model,
                                  set_coord = c_hept[v_hept[cell]],
                                  set_bool = True,
                                  set_crs = crs.uuid,
                                  title = 'v cell')
            assert v_cell.point_is_inside_xy(points_2[len(c_hept) + cell])



import os

import numpy as np
import pytest
from numpy.testing import assert_array_almost_equal

import resqpy.grid as grr
import resqpy.model as rq


def test_regular_grid_no_geometry(tmp_path):
    # issue #222

    epc = os.path.join(tmp_path, 'abstract.epc')

    model = rq.new_model(epc)

    # create a basic block grid
    grid = grr.RegularGrid(model, extent_kji = (4, 3, 2), title = 'spaced out')
    grid.create_xml(add_cell_length_properties = False)
    grid_uuid = grid.uuid

    model.store_epc()

    # check that the grid can be read
    model = rq.Model(epc)

    grid = grr.any_grid(model, uuid = grid_uuid)


def test_regular_grid_with_geometry(tmp_path):

    epc = os.path.join(tmp_path, 'concrete.epc')

    model = rq.new_model(epc)

    # create a basic block grid
    dxyz = (55.0, 65.0, 27.0)
    grid = grr.RegularGrid(model, extent_kji = (4, 3, 2), title = 'concrete', origin = (0.0, 0.0, 1000.0), dxyz = dxyz)
    grid.create_xml(add_cell_length_properties = True)
    grid_uuid = grid.uuid

    # store with constant arrays (no hdf5 data)
    model.store_epc()

    # check that the grid can be read
    model = rq.Model(epc)
    grid = grr.any_grid(model, uuid = grid_uuid)

    # check that the cell size has been preserved
    expected_dxyz_dkji = np.zeros((3, 3))
    for i in range(3):
        expected_dxyz_dkji[2 - i, i] = dxyz[i]
    assert_array_almost_equal(expected_dxyz_dkji, grid.block_dxyz_dkji)



"""Organizational object classes: features and interpretations."""

__all__ = [
    # Classes
    "BoundaryFeature",
    "EarthModelInterpretation",
    "FaultInterpretation",
    "FluidBoundaryFeature",
    "FrontierFeature",
    "GeneticBoundaryFeature",
    "GeobodyBoundaryInterpretation",
    "GeobodyFeature",
    "GeobodyInterpretation",
    "GeologicUnitFeature",
    "HorizonInterpretation",
    "OrganizationFeature",
    "RockFluidUnitFeature",
    "TectonicBoundaryFeature",
    "WellboreFeature",
    "WellboreInterpretation",
    # Functions
    "equivalent_extra_metadata",
    "alias_for_attribute",
    "extract_has_occurred_during",
    "equivalent_chrono_pairs",
    "create_xml_has_occurred_during",
]

from resqpy.organize._utils import (equivalent_extra_metadata, alias_for_attribute, extract_has_occurred_during,
                                    equivalent_chrono_pairs, create_xml_has_occurred_during)
from .boundary_feature import BoundaryFeature
from .earth_model_interpretation import EarthModelInterpretation
from .fault_interpretation import FaultInterpretation
from .fluid_boundary_feature import FluidBoundaryFeature
from .frontier_feature import FrontierFeature
from .genetic_boundary_feature import GeneticBoundaryFeature
from .geobody_boundary_interpretation import GeobodyBoundaryInterpretation
from .geobody_feature import GeobodyFeature
from .geobody_interpretation import GeobodyInterpretation
from .geologic_unit_feature import GeologicUnitFeature
from .horizon_interpretation import HorizonInterpretation
from .organization_feature import OrganizationFeature
from .rock_fluid_unit_feature import RockFluidUnitFeature
from .tectonic_boundary_feature import TectonicBoundaryFeature
from .wellbore_feature import WellboreFeature
from .wellbore_interpretation import WellboreInterpretation

# Set "module" attribute of all public objects to this path.
for _name in __all__:
    _obj = eval(_name)
    if hasattr(_obj, "__module__"):
        _obj.__module__ = __name__



"""High level extract box for well function."""

import logging

log = logging.getLogger(__name__)

import os
import numpy as np

import resqpy.crs as rqcrs
import resqpy.grid_surface as rgs
import resqpy.model as rq
import resqpy.olio.box_utilities as bx
import resqpy.olio.xml_et as rqet
import resqpy.well as rqw

from resqpy.derived_model._add_one_grid_property_array import add_one_grid_property_array
from resqpy.derived_model._extract_box import extract_box


def extract_box_for_well(epc_file = None,
                         source_grid = None,
                         min_k0 = None,
                         max_k0 = None,
                         trajectory_epc = None,
                         trajectory_uuid = None,
                         blocked_well_uuid = None,
                         column_ji0 = None,
                         column_xy = None,
                         well_name = None,
                         radius = None,
                         outer_radius = None,
                         active_cells_shape = 'tube',
                         quad_triangles = True,
                         inherit_properties = False,
                         inherit_realization = None,
                         inherit_all_realizations = False,
                         inherit_well = False,
                         set_parent_window = None,
                         new_grid_title = None,
                         new_epc_file = None):
    """Extends an existing model with a new grid extracted as an IJK box around a well trajectory in the source grid.

    arguments:
       epc_file (string): file name to rewrite the model's xml to; if source grid is None, model is loaded from this file
       source_grid (grid.Grid object, optional): if None, the epc_file is loaded and it should contain one ijk grid object
          (or one 'ROOT' grid) which is used as the source grid
       min_k0, max_k0 (integers, optional): layer range to include; default is full vertical range of source grid
       trajectory_epc (string, optional): the source file for the trajectory or blocked well, if different to that for
          the source grid
       trajectory_uuid (uuid.UUID): the uuid of the trajectory object for the well, if working from a trajectory
       blocked_well_uuid (uuid.UUID): the uuid of the blocked well object, an alternative to working from a trajectory;
          must include blocking against source_grid
       column_ji0 (integer pair, optional): an alternative to providing a trajectory: the column indices of a 'vertical' well
       column_xy (float pair, optional): an alternative to column_ji0: the x, y location used to determine the column
       well_name (string, optional): name to use for column well, ignored if trajectory_uuid is not None
       radius (float, optional): the radius around the wellbore to include in the box; units are those of grid xy values;
          radial distances are applied horizontally regardless of well inclination; if not present, only cells penetrated
          by the trajectory are included
       outer_radius (float, optional): an outer radius around the wellbore, beyond which an inactive cell mask for the
          source_grid will be set to True (inactive); units are those of grid xy values
       active_cells_shape (string, default 'tube'): the logical shape of cells marked as active in the extracted box;
          'tube' results in an active shape with circular cross section in IJ planes, that follows the trajectory; 'prism'
          activates all cells in IJ columns where any cell is within the tube; 'box' leaves the entire IJK cuboid active
       quad_triangles (boolean, default True): if True, cell K faces are treated as 4 triangles (with a common face
          centre point) when computing the intersection of the trajectory with layer interfaces (horizons); if False,
          the K faces are treated as 2 triangles
       inherit_properties (boolean, default False): if True, the new grid will have a copy of any properties associated
          with the source grid, with values taken from the extracted box
       inherit_realization (int, optional): realization number for which properties will be inherited; ignored if
          inherit_properties is False
       inherit_all_realizations (boolean, default False): if True (and inherit_realization is None), properties for all
          realizations will be inherited; if False, only properties with a realization of None are inherited; ignored if
          inherit_properties is False or inherit_realization is not None
       inherit_well (boolean, default False): if True, the new model will have a copy of the well trajectory, its crs (if
          different from that of the grid), and any related wellbore interpretation and feature
       set_parent_window (boolean, optional): if True, the extracted grid has its parent window attribute set; if False,
          the parent window is not set; if None, the default will be True if new_epc_file is None or False otherwise
       new_grid_title (string): used as the citation title text for the new grid object
       new_epc_file (string, optional): if None, the source epc_file is extended with the new grid object; if present,
          a new epc file (& associated h5 file) is created to contain the extracted grid (& crs)

    returns:
       (grid, box) where: grid is the new Grid object with extent as determined by source grid geometry, trajectory and
       radius arguments; and box is a numpy int array of shape (2, 3) with first axis covering min, max and second axis
       covering k,j,i; the box array holds the minimum and maximum indices (zero based) in the source grid that have
       been included in the extraction (nb. maximum indices are included, unlike the usual python protocol)

    notes:
       this function is designed to work fully for vertical and deviated wells; for horizontal wells use blocked well mode;
       the extracted box includes all layers between the specified min and max horizons, even if the trajectory terminates
       above the deeper horizon or does not intersect horizon(s) for other reasons;
       when specifying a column well by providing x,y the IJ column with the centre of the topmost k face closest to the
       given point is selected;
       if an outer_radius is given, a boolean property will be created for the source grid with values set True where the
       centres of the cells are beyond this distance from the well, measured horizontally; if outer_radius and new_epc_file
       are both given, the source grid will be copied to the new epc
    """

    # establish model, source grid, and trajectory model
    new_epc_file, trajectory_epc, model, traj_model, source_grid =  \
        _establish_files_and_models(epc_file, new_epc_file, trajectory_epc, source_grid)

    # ensure geometry exists as points in case of a regular (block) grid
    if source_grid.grid_representation == 'IjkBlockGrid':
        source_grid.make_regular_points_cached()

    # check arguments
    if min_k0 is None:
        min_k0 = 0
    if max_k0 is None:
        max_k0 = source_grid.nk - 1
    assert 0 <= min_k0 <= max_k0 < source_grid.nk
    assert trajectory_uuid is not None or blocked_well_uuid is not None or column_ji0 is not None
    if radius is None:
        radius = 0.0  # cells directly penetrated through a k face will still be included
    assert radius >= 0.0
    if outer_radius is not None:
        assert outer_radius >= radius
    assert active_cells_shape in ['tube', 'prism', 'box']

    # prepare cell centre points for testing inclusion
    centres = source_grid.centre_point(cache_centre_array = True)

    # initialise local variables
    box = None
    trajectory = None
    blocked_well = None
    bw_cells = None
    bw_box = None

    # establish well information, dependent on type of well object
    if trajectory_uuid is not None:
        # prepare a trajectory object
        trajectory_root = traj_model.root(obj_type = 'WellboreTrajectoryRepresentation', uuid = trajectory_uuid)
        assert trajectory_root is not None, 'trajectory object not found for uuid: ' + str(trajectory_uuid)
        trajectory = rqw.Trajectory(traj_model, uuid = trajectory_uuid)
        well_name = rqw.well_name(trajectory)
        traj_crs = rqcrs.Crs(trajectory.model, uuid = trajectory.crs_uuid)
        grid_crs = rqcrs.Crs(source_grid.model, uuid = source_grid.crs_uuid)
        # modify in-memory trajectory data to be in the same crs as grid
        traj_crs.convert_array_to(grid_crs,
                                  trajectory.control_points)  # trajectory xyz points converted in situ to grid's crs
        trajectory.crs_uuid = source_grid.crs_uuid  # note: tangent vectors might be messed up, if present
        traj_box = np.empty((2, 3))
        traj_box[0] = np.amin(trajectory.control_points, axis = 0)
        traj_box[1] = np.amax(trajectory.control_points, axis = 0)
        grid_box = source_grid.xyz_box(lazy = False)
        if not bx.boxes_overlap(traj_box, grid_box):
            log.error('no overlap of xyz boxes for trajectory and grid for trajectory uuid: ' + str(trajectory.uuid))
            return None, None
    elif blocked_well_uuid is not None:
        bw_root = traj_model.root(obj_type = 'BlockedWellboreRepresentation', uuid = blocked_well_uuid)
        assert bw_root is not None, 'blocked well object not found for uuid: ' + str(blocked_well_uuid)
        blocked_well = rqw.BlockedWell(traj_model, uuid = blocked_well_uuid)
        well_name = rqw.well_name(blocked_well)
        bw_box = blocked_well.box(grid_uuid = source_grid.uuid)
        assert bw_box is not None, 'blocked well does not include cells in source grid'
        assert bw_box[0, 0] <= max_k0 and bw_box[1, 0] >= min_k0,  \
            'blocked well does not include cells in specified layer range'
        bw_cells = blocked_well.cell_indices_for_grid_uuid(source_grid.uuid)
    else:
        column_ji0, well_name = _check_column(source_grid, column_ji0, column_xy, well_name)

    # create cell mask
    inclusion_mask, outer_inactive_mask = _build_cell_masks(source_grid, centres, min_k0, max_k0, trajectory,
                                                            blocked_well, bw_cells, column_ji0, quad_triangles, radius,
                                                            outer_radius)

    # derive box from inclusion mask
    box = _box_from_inclusion_mask(source_grid, inclusion_mask, min_k0, max_k0)

    # prepare inactive mask to merge in for new grid
    box_inactive = _make_inactive_mask(active_cells_shape, inclusion_mask, box)

    # establish title for the new grid
    new_grid_title = _invent_title(new_grid_title, trajectory, blocked_well, column_ji0)

    # perform the main grid extraction
    grid = extract_box(epc_file,
                       source_grid = source_grid,
                       box = box,
                       box_inactive = box_inactive,
                       inherit_properties = inherit_properties,
                       inherit_realization = inherit_realization,
                       inherit_all_realizations = inherit_all_realizations,
                       set_parent_window = set_parent_window,
                       new_grid_title = new_grid_title,
                       new_epc_file = new_epc_file)

    # inherit well if requested
    if inherit_well and new_epc_file:
        _inherit_well(new_epc_file, grid, traj_model, well_name, trajectory, blocked_well, column_ji0, box)

    # add mask property for outer radius, if specified
    if outer_radius is not None:
        _add_outer_mask_property(epc_file, new_epc_file, outer_inactive_mask, source_grid, well_name)

    return grid, box


def _box_from_inclusion_mask(source_grid, inclusion_mask, min_k0, max_k0):
    min_j0 = 0
    while min_j0 < source_grid.nj - 1 and not np.any(inclusion_mask[:, min_j0, :]):
        min_j0 += 1
    max_j0 = source_grid.nj - 1
    while max_j0 > 0 and not np.any(inclusion_mask[:, max_j0, :]):
        max_j0 -= 1
    assert max_j0 >= min_j0
    min_i0 = 0
    while min_i0 < source_grid.ni - 1 and not np.any(inclusion_mask[:, :, min_i0]):
        min_i0 += 1
    max_i0 = source_grid.ni - 1
    while max_i0 > 0 and not np.any(inclusion_mask[:, :, max_i0]):
        max_i0 -= 1
    assert max_i0 >= min_i0
    box = np.array([[min_k0, min_j0, min_i0], [max_k0, max_j0, max_i0]], dtype = int)
    log.info('box for well is: ' + bx.string_iijjkk1_for_box_kji0(box) + ' (simulator protocol)')
    return box


def _build_cell_masks(source_grid, centres, min_k0, max_k0, trajectory, blocked_well, bw_cells, column_ji0,
                      quad_triangles, radius, outer_radius):
    # todo: handle base interfaces above k gaps
    # either work with grid layers of interfaces between layers (grid horizons)
    # intialise masks to False
    radius_sqr = radius * radius
    outer_radius_sqr = None if outer_radius is None else outer_radius * outer_radius
    inclusion_mask = np.zeros(source_grid.extent_kji, dtype = bool)
    outer_inactive_mask = None if outer_radius is None else np.zeros(source_grid.extent_kji, dtype = bool)
    h_or_l = 'layer' if trajectory is None else 'horizon'
    end_k0 = max_k0 + 1 if trajectory is None else max_k0 + 2
    warned = False
    for k in range(min_k0, end_k0):
        cols, intersect_points = _find_intersections(source_grid, k, trajectory, blocked_well, bw_cells, column_ji0,
                                                     centres, quad_triangles)
        if cols is None or len(cols) == 0:
            if not warned:
                log.warning(f"no intersection found between well and {h_or_l}(s) such as: {k}")
                warned = True
            continue
        count = cols.shape[0]
        assert len(intersect_points) == count
        if count > 1:
            log.warning(f"{count} intersections found between well and {h_or_l}: {k}")
        layer_mask = np.zeros((source_grid.nj, source_grid.ni), dtype = bool)  # to be set True within radius
        if outer_radius is not None:
            # to be set False within outer_radius
            outer_layer_mask = np.ones((source_grid.nj, source_grid.ni), dtype = bool)
        for intersect in range(count):
            log.debug(f"well intersects {h_or_l} {k} in column j0,i0: {cols[intersect, 0]}, {cols[intersect, 1]}")
            if radius > 0.0 or outer_radius is not None:
                if k < source_grid.nk:
                    vectors = centres[k] - intersect_points[intersect].reshape((1, 1, 3))
                    distance_sqr = vectors[..., 0] * vectors[..., 0] + vectors[..., 1] * vectors[..., 1]
                    if radius > 0.0:
                        layer_mask = np.logical_or(layer_mask, np.less_equal(distance_sqr, radius_sqr))
                    if outer_radius is not None:
                        outer_layer_mask = np.logical_and(outer_layer_mask,
                                                          np.greater_equal(distance_sqr, outer_radius_sqr))
                if k > 0 and (not source_grid.k_gaps or k >= source_grid.nk - 1 or
                              not source_grid.k_gap_after_array[k - 1]):
                    vectors = centres[k - 1] - intersect_points[intersect].reshape((1, 1, 3))
                    distance_sqr = vectors[..., 0] * vectors[..., 0] + vectors[..., 1] * vectors[..., 1]
                    if radius > 0.0:
                        layer_mask = np.logical_or(layer_mask, np.less_equal(distance_sqr, radius_sqr))
                    if outer_radius is not None:
                        outer_layer_mask = np.logical_and(outer_layer_mask,
                                                          np.greater_equal(distance_sqr, outer_radius_sqr))
            layer_mask[cols[intersect, 0], cols[intersect, 1]] = True
        if k <= max_k0:
            inclusion_mask[k] = layer_mask
            if outer_radius is not None:
                outer_inactive_mask[k] = outer_layer_mask
        if k > min_k0:
            inclusion_mask[k - 1] = np.logical_or(inclusion_mask[k - 1], layer_mask)
            if outer_radius is not None:
                outer_inactive_mask[k - 1] = np.logical_and(outer_inactive_mask[k - 1], outer_layer_mask)
        log.debug(f"number of columns found in {h_or_l} {k} within radius around well: {np.count_nonzero(layer_mask)}")
    inc_count = np.count_nonzero(inclusion_mask)
    if inc_count == 0:
        log.error('no cells found within search radius around well')
        return None, None
    log.info('total number of cells found within radius around well: ' + str(inc_count))
    return inclusion_mask, outer_inactive_mask


def _invent_title(new_grid_title, trajectory, blocked_well, column_ji0):
    if not new_grid_title:
        if trajectory is not None:
            new_grid_title = 'local grid extracted for well: ' + str(trajectory.title)
        elif blocked_well is not None:
            new_grid_title = 'local grid extracted for blocked well: ' + str(blocked_well.title)
        elif column_ji0 is not None:
            new_grid_title = 'local grid extracted around column i, j (1 based): ' +  \
                             str(column_ji0[1] + 1) + ', ' + str(column_ji0[0] + 1)
        else:  # should not happen
            new_grid_title = 'local grid extracted for well'
    return new_grid_title


def _add_outer_mask_property(epc_file, new_epc_file, outer_inactive_mask, source_grid, well_name):
    if new_epc_file:
        # TODO: copy source grid and reassign source_grid to new copy
        outer_epc = new_epc_file
    else:
        outer_epc = epc_file
    # todo: make local property kind, or reuse active local property kind?
    add_one_grid_property_array(outer_epc,
                                outer_inactive_mask,
                                'discrete',
                                grid_uuid = source_grid.uuid,
                                source_info = 'extract box for well outer radius',
                                title = 'distant mask for well ' + str(well_name),
                                discrete = True)


def _inherit_well(new_epc_file, grid, traj_model, well_name, trajectory, blocked_well, column_ji0, box):
    newer_model = rq.Model(new_epc_file)
    if trajectory is None and blocked_well is None:
        log.info('creating well objects for column')
        box_column_ji0 = (column_ji0[0] - box[0, 1], column_ji0[1] - box[0, 2])
        bw = rqw.BlockedWell(newer_model,
                             grid = grid,
                             column_ji0 = box_column_ji0,
                             well_name = well_name,
                             use_face_centres = True)
        bw.write_hdf5(create_for_trajectory_if_needed = True)
        bw.create_xml(create_for_trajectory_if_needed = True, title = well_name)
    elif blocked_well is not None:
        log.info('inheriting trajectory for blocked well')  # recursively copies referenced parts
        newer_model.copy_part_from_other_model(traj_model, traj_model.part(uuid = blocked_well.trajectory.uuid))
    else:
        log.info('inheriting well trajectory')  # recursively copies referenced parts
        newer_model.copy_part_from_other_model(traj_model, traj_model.part(uuid = trajectory.uuid))
    newer_model.h5_release()
    newer_model.store_epc()


def _make_inactive_mask(active_cells_shape, inclusion_mask, box):
    if active_cells_shape in ['tube', 'prism']:
        if active_cells_shape == 'prism':
            layer_mask = np.any(inclusion_mask, axis = 0)
            inclusion_mask[:] = layer_mask
        box_inactive = np.logical_not(inclusion_mask[box[0, 0]:box[1, 0] + 1, box[0, 1]:box[1, 1] + 1,
                                                     box[0, 2]:box[1, 2] + 1])
    else:  # 'box' option: leave all cells active (except where inactive in source grid)
        box_inactive = None
    return box_inactive


def _check_column(source_grid, column_ji0, column_xy, well_name):
    if column_ji0 is None:
        assert len(column_xy) == 2
        column_ji0 = source_grid.find_cell_for_point_xy(column_xy[0], column_xy[1])
        if column_ji0[0] is None or column_ji0[1] is None:
            log.error('no column found for x, y: ' + str(column_xy[0]) + ', ' + str(column_xy[1]))
        return None, None
    assert len(column_ji0) == 2
    assert 0 <= column_ji0[0] < source_grid.nj and 0 <= column_ji0[1] < source_grid.ni
    if not well_name:
        well_name = 'well for global column ' + str(column_ji0[1] + 1) + ', ' + str(column_ji0[0] + 1)
    return column_ji0, well_name


def _establish_files_and_models(epc_file, new_epc_file, trajectory_epc, source_grid):
    assert epc_file or new_epc_file, 'epc file name not specified'
    if new_epc_file and epc_file and (
        (new_epc_file == epc_file) or
        (os.path.exists(new_epc_file) and os.path.exists(epc_file) and os.path.samefile(new_epc_file, epc_file))):
        new_epc_file = None
    assert epc_file or source_grid is not None, 'neither epc file name nor source grid supplied'
    if trajectory_epc == epc_file:
        trajectory_epc = None
    if epc_file:
        model = rq.Model(epc_file)
        if source_grid is None:
            source_grid = model.grid()  # requires there to be exactly one grid in model (or one named ROOT)
    else:
        model = source_grid.model
    assert source_grid.grid_representation in ['IjkGrid', 'IjkBlockGrid']
    assert model is not None
    if trajectory_epc is None:
        traj_model = model
    else:
        traj_model = rq.Model(trajectory_epc)
    assert traj_model is not None
    return new_epc_file, trajectory_epc, model, traj_model, source_grid


def _find_intersections(source_grid, k, trajectory, blocked_well, bw_cells, column_ji0, centres, quad_triangles):
    if trajectory is None:
        if blocked_well is None:
            cols_ji0 = np.array(column_ji0, dtype = int).reshape((1, 2))
            cols, intersect_points = cols_ji0, centres[k, column_ji0[0], column_ji0[1]].reshape((1, 3))
        else:
            selected_cells = np.where(bw_cells[:, 0] == k)[0]
            cells = bw_cells[selected_cells]
            cols = cells[:, 1:]
            intersect_points = centres[cells[:, 0], cells[:, 1], cells[:, 2]]
    else:
        if k < source_grid.nk:
            cols, intersect_points = rgs.find_intersections_of_trajectory_with_layer_interface(
                trajectory, source_grid, k0 = k, ref_k_faces = 'top', quad_triangles = quad_triangles)
        else:
            cols, intersect_points = rgs.find_intersections_of_trajectory_with_layer_interface(
                trajectory, source_grid, k0 = k - 1, ref_k_faces = 'base', quad_triangles = quad_triangles)
    return cols, intersect_points



import os

import numpy as np
import pytest

import resqpy.crs as rqc
import resqpy.grid as grr
import resqpy.model as rq
import resqpy.olio.uuid as bu
import resqpy.olio.write_hdf5 as rwh5
import resqpy.olio.xml_et as rqet
import resqpy.property as rqp
import resqpy.well as rqw


def test_model(tmp_path):

    epc = os.path.join(tmp_path, 'model.epc')
    model = rq.new_model(epc)
    assert model is not None
    crs = rqc.Crs(model)
    crs_root = crs.create_xml()
    model.store_epc()
    assert os.path.exists(epc)
    md_datum_1 = rqw.MdDatum(model, location = (0.0, 0.0, -50.0), crs_uuid = crs.uuid)
    md_datum_1.create_xml(title = 'Datum & 1')
    md_datum_2 = rqw.MdDatum(model, location = (3.0, 0.0, -50.0), crs_uuid = crs.uuid)
    md_datum_2.create_xml(title = 'Datum < 2')
    assert len(model.uuids(obj_type = 'MdDatum')) == 2
    model.store_epc()

    model = rq.Model(epc)
    assert model is not None
    assert len(model.uuids(obj_type = 'MdDatum')) == 2
    datum_part_1 = model.part(obj_type = 'MdDatum', title = '1', title_mode = 'ends')
    datum_part_2 = model.part(obj_type = 'MdDatum', title = '2', title_mode = 'ends')
    assert datum_part_1 is not None and datum_part_2 is not None and datum_part_1 != datum_part_2
    datum_uuid_1 = rqet.uuid_in_part_name(datum_part_1)
    datum_uuid_2 = rqet.uuid_in_part_name(datum_part_2)
    assert not bu.matching_uuids(datum_uuid_1, datum_uuid_2)
    p1 = model.uuid_part_dict[bu.uuid_as_int(datum_uuid_1)]
    p2 = model.uuid_part_dict[bu.uuid_as_int(datum_uuid_2)]
    assert p1 == datum_part_1 and p2 == datum_part_2


def test_model_iterators(example_model_with_well):

    model, well_interp, datum, traj = example_model_with_well

    w = next(model.iter_wellbore_interpretations())
    d = next(model.iter_md_datums())
    t = next(model.iter_trajectories())

    assert bu.matching_uuids(w.uuid, well_interp.uuid)
    assert bu.matching_uuids(d.uuid, datum.uuid)
    assert bu.matching_uuids(t.uuid, traj.uuid)
    assert w == well_interp
    assert d == datum
    assert t == traj


def test_model_iterate_objects(example_model_with_well):

    from resqpy.organize import WellboreFeature, WellboreInterpretation

    model, well_interp, _, _ = example_model_with_well

    # Try iterating over wellbore features

    wells = model.iter_objs(WellboreFeature)
    wells = list(wells)
    w = wells[0]
    assert len(wells) == 1
    assert isinstance(wells[0], WellboreFeature)
    assert wells[0].title == "well A"

    # Try iterating over wellbore interpretations

    interps = model.iter_objs(WellboreInterpretation)
    interps = list(interps)
    assert len(interps) == 1
    assert isinstance(interps[0], WellboreInterpretation)
    assert interps[0] == well_interp


def test_model_iter_crs(example_model_and_crs):

    model, crs_1 = example_model_and_crs

    crs_list = list(model.iter_crs())
    assert len(crs_list) == 1

    crs_2 = crs_list[0]
    assert crs_2 == crs_1


def test_model_iter_crs_empty(tmp_model):

    # Should raise an exception if no CRS exists
    with pytest.raises(StopIteration):
        next(tmp_model.iter_crs())


def test_model_as_graph(example_model_with_well):

    model, well_interp, datum, traj = example_model_with_well
    crs = next(model.iter_crs())

    nodes, edges = model.as_graph()

    # Check nodes
    for obj in [well_interp, datum, traj, crs]:
        assert str(obj.uuid) in nodes.keys()
        if hasattr(obj, 'title'):  # TODO: remove this when all objects have this attribute
            assert obj.title == nodes[str(obj.uuid)]['title']

    # Check edges
    assert frozenset([str(traj.uuid), str(datum.uuid)]) in edges
    assert frozenset([str(traj.uuid), str(crs.uuid)]) in edges
    assert frozenset([str(traj.uuid), str(well_interp.uuid)]) in edges
    assert frozenset([str(datum.uuid), str(traj.uuid)]) in edges

    # Test uuid subset
    nodes, edges = model.as_graph(uuids_subset = [datum.uuid])
    assert len(nodes.keys()) == 1
    assert len(edges) == 0


def test_model_copy_all_parts(example_model_with_properties):

    epc = example_model_with_properties.epc_file
    dir = example_model_with_properties.epc_directory
    copied_epc = os.path.join(dir, 'copied.epc')

    # test copying without consolidation
    original = rq.Model(epc)
    assert original is not None
    copied = rq.new_model(copied_epc)
    copied.copy_all_parts_from_other_model(original, consolidate = False)

    assert set(original.uuids()) == set(copied.uuids())
    assert set(original.parts()) == set(copied.parts())

    # test without consolidation of two crs objects
    copied = rq.new_model(copied_epc)
    new_crs = rqc.Crs(copied)
    new_crs.create_xml()

    copied.copy_all_parts_from_other_model(original, consolidate = False)

    assert len(copied.parts()) == len(original.parts()) + 1
    assert set(original.parts()).issubset(set(copied.parts()))
    assert len(copied.parts(obj_type = 'LocalDepth3dCrs')) == 2

    # test with consolidation of two crs objects
    copied = rq.new_model(copied_epc)
    new_crs = rqc.Crs(copied)
    new_crs.create_xml()

    copied.copy_all_parts_from_other_model(original, consolidate = True)

    assert len(copied.parts()) == len(original.parts())
    assert len(copied.parts(obj_type = 'LocalDepth3dCrs')) == 1

    crs_uuid = copied.uuid(obj_type = 'LocalDepth3dCrs')
    assert (bu.matching_uuids(crs_uuid, new_crs.uuid) or
            bu.matching_uuids(crs_uuid, original.uuid(obj_type = 'LocalDepth3dCrs')))

    # test write and re-load of copied model
    copied.store_epc()
    re_opened = rq.Model(copied_epc)
    assert re_opened is not None

    assert len(copied.parts()) == len(original.parts())

    crs_uuid = re_opened.uuid(obj_type = 'LocalDepth3dCrs')
    assert (bu.matching_uuids(crs_uuid, new_crs.uuid) or
            bu.matching_uuids(crs_uuid, original.uuid(obj_type = 'LocalDepth3dCrs')))


def test_model_copy_all_parts_non_resqpy_hdf5_paths(example_model_with_properties):
    # this test uses low level methods to override the hdf5 internal paths for some new objects

    epc = example_model_with_properties.epc_file
    dir = example_model_with_properties.epc_directory
    copied_epc = os.path.join(dir, 'copied.epc')
    extent_kji = (3, 5, 5)  # needs to match extent of grid in example model

    # add some properties with bespoke internal hdf5 paths
    original = rq.Model(epc)
    assert original is not None
    grid_uuid = original.uuid(obj_type = 'IjkGridRepresentation')
    assert grid_uuid is not None
    data = np.linspace(0.0, 1000.0, num = 75).reshape(extent_kji)

    hdf5_paths = [
        '/RESQML/uuid_waffle/values', '/RESQML/waffle_uuid/unusual_name', 'RESQML/class_uuid_waffle/values0',
        'RESQML/something_interesting/array', '/RESQML/abrupt', 'no_resqml/uuid/values'
    ]

    prop_list = []
    path_list = []  # elements will have 'uuid' replaced with uuid string
    h5_reg = rwh5.H5Register(original)
    for i, hdf5_path in enumerate(hdf5_paths):
        prop = rqp.Property(original, support_uuid = grid_uuid)
        if 'uuid' in hdf5_path:
            path = hdf5_path.replace('uuid', str(prop.uuid))
        else:
            path = hdf5_path
        prop.prepare_import(cached_array = data,
                            source_info = 'test data',
                            keyword = f'TEST{i}',
                            property_kind = 'continuous',
                            uom = 'm')
        for entry in prop.collection.imported_list:  # only one entry
            # override internal hdf5 path
            h5_reg.register_dataset(entry[0], 'values_patch0', prop.collection.__dict__[entry[3]], hdf5_path = path)
        prop_list.append(prop)
        path_list.append(path)
    h5_reg.write(mode = 'a')

    for prop, hdf5_path in zip(prop_list, path_list):
        root_node = prop.create_xml(find_local_property_kind = False)
        assert root_node is not None
        # override the hdf5 internal path in the xml tree
        path_node = rqet.find_nested_tags(root_node, ['PatchOfValues', 'Values', 'Values', 'PathInHdfFile'])
        assert path_node is not None
        path_node.text = hdf5_path

    # rewrite model
    original.store_epc()

    # re-open model and test copying
    original = rq.Model(epc)
    assert original is not None
    copied = rq.new_model(copied_epc)
    copied.copy_all_parts_from_other_model(original, consolidate = False)

    assert set(original.uuids()) == set(copied.uuids())
    assert set(original.parts()) == set(copied.parts())


def test_model_context(tmp_path):

    # Create a new model
    epc_path = str(tmp_path / 'tmp_model.epc')
    model = rq.new_model(epc_path)
    crs = rqc.Crs(parent_model = model, title = 'kuzcotopia')
    crs_uuid = crs.uuid
    crs.create_xml()
    model.store_epc()
    del crs, model

    # Re-open model in read/write mode
    with rq.ModelContext(epc_path, mode = "rw") as model2:

        crs2 = rqc.Crs(model2, uuid = crs_uuid)
        assert len(list(model2.iter_crs())) == 1
        assert crs2.title == 'kuzcotopia'

        # Make a change
        crs2.title = 'wabajam'
        crs2.create_xml(reuse = False)

    # Re-open model in read mode
    with rq.ModelContext(epc_path, mode = "r") as model3:

        # Check model has loaded correctly
        assert len(list(model3.iter_crs())) == 1
        crs3 = rqc.Crs(model3, uuid = crs_uuid)
        assert crs3.title == 'wabajam'

    # Overwrite model
    with rq.ModelContext(epc_path, mode = "create") as model4:
        # Should be empty
        crs_list = list(model4.iter_crs())
        assert len(crs_list) == 0


def test_multiple_epc_sharing_one_hdf5(tmp_path, example_model_with_prop_ts_rels):

    # get some key data from the full model
    full_model = example_model_with_prop_ts_rels
    full_epc_path = full_model.epc_file
    full_count = full_model.number_of_parts()
    assert full_count > 4
    hdf5_path = full_model.h5_file_name(override = 'none')
    assert os.path.exists(hdf5_path)
    ext_uuid = full_model.h5_uuid()
    assert ext_uuid is not None
    grid_uuid = full_model.uuid(obj_type = 'IjkGridRepresentation')
    assert grid_uuid is not None
    grid = full_model.grid(uuid = grid_uuid)
    pc = grid.property_collection
    assert pc is not None and pc.number_of_parts() > 7
    ts_uuid = full_model.uuid(obj_type = 'TimeSeries')
    assert ts_uuid is not None
    discrete_prop_uuid_list = full_model.uuids(obj_type = 'DiscreteProperty')
    assert len(discrete_prop_uuid_list) == 3

    # create a couple of sub model file names, one for each of two realisations
    epc_0 = os.path.join(tmp_path, 'model_0.epc')
    epc_1 = os.path.join(tmp_path, 'model_1.epc')

    # for each sub model...
    for i, epc_path in enumerate((epc_0, epc_1)):

        # create a new, empty, model without the usual new hdf5 file and external part
        model = rq.Model(epc_file = epc_path, new_epc = True, create_basics = True, create_hdf5_ext = False)

        # switch off the default hdf5 filename override
        model.h5_set_default_override('none')

        # create an hdf5 external part referring to the full model's hdf5 file, preserving the ext uuid
        ext_node = model.create_hdf5_ext(file_name = hdf5_path, uuid = ext_uuid)
        assert ext_node is not None

        # check that the correct hdf5 path will be returned for the sub model when not overriding
        sub_hdf5_path = model.h5_file_name(override = 'none')
        assert sub_hdf5_path, 'hdf5 path not established for sub model'
        assert os.path.samefile(hdf5_path, sub_hdf5_path)

        # copy some common parts into the sub model (will include referenced crs)
        for uuid in [grid_uuid, ts_uuid] + discrete_prop_uuid_list:
            model.copy_part_from_other_model(full_model,
                                             full_model.part(uuid = uuid),
                                             self_h5_file_name = hdf5_path,
                                             other_h5_file_name = hdf5_path)

        # copy some realisation specific properties
        for pk in ('net to gross ratio', 'porosity'):
            property_part = pc.singleton(property_kind = pk, realization = i)
            assert property_part is not None
            model.copy_part_from_other_model(full_model,
                                             property_part,
                                             self_h5_file_name = hdf5_path,
                                             other_h5_file_name = hdf5_path)

        # save the epc for the sub model
        model.store_epc()

        # check the number of parts in the sub model is as expected (includes crs and ext)
        assert model.number_of_parts() == 6 + len(discrete_prop_uuid_list)

    # re-open all 3 models and see how they shape up
    full_model = rq.Model(full_epc_path)
    model_0 = rq.Model(epc_0)
    model_1 = rq.Model(epc_1)
    assert full_model.number_of_parts() == full_count
    full_uuids = full_model.uuids()
    assert len(full_uuids) == full_count - 1  # adjusting for ext part which is not included in uuids()

    # for each sub model...
    for model in (model_0, model_1):
        # switch off hdf5 filename override
        model.h5_set_default_override('none')
        # check that the number of parts in the sub model is still as expected
        assert model.number_of_parts() == 6 + len(discrete_prop_uuid_list)
        # check that all sub model uuids exist in the full model, and have the same type
        for uuid in model.uuids():
            found = False
            for full_uuid in full_uuids:
                if bu.matching_uuids(uuid, full_uuid):
                    found = True
                    break
            assert found, 'new uuid found in sub model'
            assert model.type_of_uuid(uuid) == full_model.type_of_uuid(uuid)

    # establish some grid objects with property collections
    grid = full_model.grid()
    grid_0 = model_0.grid()
    assert grid == grid_0, 'grid equivalence based on uuid failed'
    grid_1 = model_1.grid()
    assert grid == grid_1, 'grid equivalence based on uuid failed'

    # check that realisation-specific properties are different in the two sub models
    for pk in ('net to gross ratio', 'porosity'):
        prop_0 = grid_0.property_collection.single_array_ref(property_kind = pk)
        assert prop_0 is not None
        prop_1 = grid_1.property_collection.single_array_ref(property_kind = pk)
        assert prop_1 is not None
        assert not np.any(np.isclose(prop_0, prop_1))


def test_one_epc_using_multiple_hdf5(tmp_path, example_model_with_prop_ts_rels):

    model = example_model_with_prop_ts_rels
    epc = model.epc_file
    model.h5_set_default_override('dir')
    grid = model.grid()
    assert grid is not None and grid.property_collection is not None
    pc = grid.property_collection

    # add a couple of new hdf5 parts
    jitter_ext_files = []
    jitter_ext_nodes = []
    jitter_ext_uuids = []
    for i in range(2):
        filename = f'jitter_{i}.h5'
        jitter_ext_files.append(filename)
        jitter_ext_nodes.append(model.create_hdf5_ext(file_name = filename))
        jitter_ext_uuids.append(bu.uuid_from_string(jitter_ext_nodes[-1].attrib['uuid']))
    assert None not in jitter_ext_nodes and None not in jitter_ext_uuids

    # generate a couple of new permeability arrays
    k_original = pc.single_array_ref(property_kind = 'rock permeability', indexable = 'cells')
    assert k_original is not None
    k_new = []
    for i in range(2):
        k_new.append(k_original + 1.0 + 5.0 * float(i) + np.random.random(k_original.shape))

    # add the new arrays to the property collection but writing to the new, distinct hdf5 files
    for i, (k_a, hdf5_file, ext_uuid) in enumerate(zip(k_new, jitter_ext_files, jitter_ext_uuids)):
        pc.add_cached_array_to_imported_list(k_a,
                                             source_info = 'testing multi hdf5',
                                             keyword = 'Jittery Permeability',
                                             discrete = False,
                                             uom = 'mD',
                                             property_kind = 'permeability rock',
                                             facet_type = 'direction',
                                             facet = 'IJK',
                                             realization = i,
                                             indexable_element = 'cells')
        pc.write_hdf5_for_imported_list(file_name = jitter_ext_files[i])
        pc.create_xml_for_imported_list_and_add_parts_to_model(ext_uuid = ext_uuid)

    # store the updated epc
    model.store_epc()

    # check that the new hdf5 files exist and are the expected size
    for filename in jitter_ext_files:
        file_path = os.path.join(model.epc_directory, filename)
        assert os.path.exists(file_path)
        assert os.path.getsize(file_path) >= 8 * grid.cell_count()

    # re-open the model and check that multiple ext parts are present
    model = rq.Model(epc)
    assert len(model.external_parts_list()) == model.parts_count_by_type('EpcExternalPartReference')[0][1] == 3

    # check that we can access the stored arrays
    model.h5_set_default_override('none')
    grid = model.grid()
    assert grid is not None
    pc = grid.property_collection
    assert pc is not None
    jitter_pc = rqp.selective_version_of_collection(pc, citation_title = 'Jittery Permeability')
    assert jitter_pc.number_of_parts() == 2
    k_0 = jitter_pc.single_array_ref(realization = 0)
    k_1 = jitter_pc.single_array_ref(realization = 1)
    assert k_0 is not None and k_1 is not None
    assert np.mean(k_1) >= 4.0 + np.mean(k_0)


def test_root_for_time_series(example_model_with_prop_ts_rels):
    model = example_model_with_prop_ts_rels
    ts_root = model.root_for_time_series()
    assert ts_root is not None
    assert rqet.node_type(ts_root, strip_obj = True) == 'TimeSeries'


def test_grid_list(example_model_and_crs):
    model, crs = example_model_and_crs
    # create some grid objects
    grid_a = grr.RegularGrid(model, extent_kji = (2, 2, 2), crs_uuid = crs.uuid, title = 'GRID A')
    grid_a.write_hdf5()
    grid_a.create_xml(write_active = False, add_cell_length_properties = False)
    grid_b = grr.RegularGrid(model, extent_kji = (3, 3, 3), crs_uuid = crs.uuid, title = 'GRID B')
    grid_b.write_hdf5()
    grid_b.create_xml(write_active = False, add_cell_length_properties = False)
    grid_c = grr.RegularGrid(model, extent_kji = (4, 4, 4), crs_uuid = crs.uuid, title = 'GRID C')
    grid_c.write_hdf5()
    grid_c.create_xml(write_active = False, add_cell_length_properties = False)
    # access a grid part
    grid_1 = model.grid(title = 'GRID C')
    assert grid_1 is not None
    assert grid_1 == grid_c, 'grid comparison based on uuids failed'
    assert len(model.parts(obj_type = 'IjkGridRepresentation')) == 3
    # check that the call to grid() has added the returned grid to the cache list
    assert len(model.grid_list_uuid_list()) == len(model.grid_list) == 1
    assert model.grid_list[0] == grid_c, 'grid comparison based on uuids failed'
    assert model.grid_list[0] is grid_1
    # access another grid by uuid and check that it is added to cached list
    grid_2 = model.grid(uuid = grid_a.uuid)
    assert grid_2 == grid_a
    assert len(model.grid_list_uuid_list()) == len(model.grid_list) == 2
    # add all 3 grids to the grid cache list, checking for duplicates
    model.add_grid(grid_a, check_for_duplicates = True)
    model.add_grid(grid_b, check_for_duplicates = True)
    model.add_grid(grid_c, check_for_duplicates = True)
    assert len(model.parts(obj_type = 'IjkGridRepresentation')) == 3
    assert len(model.grid_list_uuid_list()) == 3
    # check use of cached grids
    grid_3a = model.grid_for_uuid_from_grid_list(grid_b.uuid)
    assert grid_3a == grid_b, 'grid comparison based on uuids failed'
    assert grid_3a is grid_b
    grid_3b = model.grid_for_uuid_from_grid_list(grid_b.uuid)
    assert grid_3a is grid_3b
    assert tuple(grid_3a.extent_kji) == (3, 3, 3)



import os

import pytest

import resqpy.model as rq
import resqpy.time_series as rqts


def test_merge_timeseries():
    model = rq.Model(create_basics = True)

    timestamps1 = ['2020-01-01T00:00:00Z', '2015-01-01T00:00:00Z']
    timestamps2 = ['2021-01-01T00:00:00Z', '2014-01-01T00:00:00Z']

    timeseries1 = rqts.time_series_from_list(timestamps1, parent_model = model)
    timeseries1.set_model(model)
    timeseries1.create_xml()
    timeseries2 = rqts.time_series_from_list(timestamps2, parent_model = model)
    timeseries2.set_model(model)
    timeseries2.create_xml()

    sortedtimestamps = sorted(timeseries1.datetimes() + timeseries2.datetimes())

    timeseries_uuids = (timeseries1.uuid, timeseries2.uuid)

    newts, newtsuuid, timeserieslist = rqts.merge_timeseries_from_uuid(model, timeseries_uuids)

    assert len(newts.timestamps) == len(timeseries1.timestamps) + len(timeseries2.timestamps)

    for idx, timestamp in enumerate(newts.datetimes()):
        assert timestamp == sortedtimestamps[idx]

    # Now test duplication doesn't create duplicate timestamps during merge, I want a unique set of merged timestamps

    timestamps3 = [timestamp for timestamp in timestamps1]
    timeseries3 = rqts.time_series_from_list(timestamps3, parent_model = model)
    timeseries3.set_model(model)
    timeseries3.create_xml()

    timeseries_uuids = (timeseries1.uuid, timeseries2.uuid, timeseries3.uuid)

    newts2, _, _ = rqts.merge_timeseries_from_uuid(model, timeseries_uuids)

    assert len(newts.timestamps) == len(newts2.timestamps)

    for idx, timestamp in enumerate(newts2.datetimes()):
        assert timestamp == sortedtimestamps[idx]

    return True


def test_time_series_from_list(tmp_path):
    epc = os.path.join(tmp_path, 'ts_list.epc')
    model = rq.new_model(epc)
    ts_list = ['2022-01-01', '2022-02-01', '2022-03-01', '2022-04-01']
    ts = rqts.time_series_from_list(ts_list, parent_model = model)
    assert ts.number_of_timestamps() == 4
    assert ts.days_between_timestamps(0, 3) == 31 + 28 + 31
    ts.create_xml()
    model.store_epc()
    model = rq.Model(epc)
    ts_uuid = model.uuid(obj_type = 'TimeSeries')
    assert ts_uuid is not None
    ts = rqts.any_time_series(model, uuid = ts_uuid)
    assert isinstance(ts, rqts.TimeSeries)
    assert ts.number_of_timestamps() == 4
    assert ts.days_between_timestamps(0, 3) == 31 + 28 + 31
    assert ts.timeframe == 'human'


def test_time_series_from_args(tmp_path):
    epc = os.path.join(tmp_path, 'ts_args.epc')
    model = rq.new_model(epc)
    ts = rqts.TimeSeries(model,
                         first_timestamp = '1963-08-23',
                         daily = 8,
                         monthly = 4,
                         quarterly = 8,
                         yearly = 5,
                         title = 'late 60s')
    assert ts.number_of_timestamps() == 26
    assert ts.days_between_timestamps(2, 3) == 1
    assert ts.days_between_timestamps(0, 25) == 8 + 4 * 30 + 8 * 90 + 5 * 365
    ts.create_xml()
    model.store_epc()
    model = rq.Model(epc)
    ts_uuid = model.uuid(obj_type = 'TimeSeries')
    assert ts_uuid is not None
    ts = rqts.TimeSeries(model, uuid = ts_uuid)
    assert ts.number_of_timestamps() == 26
    assert ts.days_between_timestamps(2, 3) == 1
    assert ts.days_between_timestamps(0, 25) == 8 + 4 * 30 + 8 * 90 + 5 * 365


def test_geologic_time_series(tmp_path):
    epc = os.path.join(tmp_path, 'ts_geologic.epc')
    model = rq.new_model(epc)
    # Cretaceous Age start times, in Ma, with random use of sign to check it is ignored
    ma_list = (145, 72.1, -83.6, 86.3, 89.8, 93.9, 100.5, -113, -125, 129.4, 132.9, 139.8)
    ts_list = [int(round(ma * 1000000)) for ma in ma_list]
    ts_list_2 = [int(round(ma * 2000000)) for ma in ma_list]
    ts = rqts.time_series_from_list(ts_list, parent_model = model)
    assert ts.number_of_timestamps() == 12
    ts.create_xml()
    ts_2 = rqts.GeologicTimeSeries.from_year_list(model, year_list = ts_list_2, title = 'using class method')
    ts_2.create_xml()
    model.store_epc()
    model = rq.Model(epc)
    ts_uuids = model.uuids(obj_type = 'TimeSeries')
    assert ts_uuids is not None and len(ts_uuids) == 2
    for ts_uuid in ts_uuids:
        ts = rqts.any_time_series(model, uuid = ts_uuid)
        assert isinstance(ts, rqts.GeologicTimeSeries)
        assert ts.timeframe == 'geologic'
        assert ts.number_of_timestamps() == 12
        assert ((ts.timestamps[0] == -145000000 and ts.timestamps[-1] == -72100000) or
                (ts.timestamps[0] == -145000000 * 2 and ts.timestamps[-1] == -72100000 * 2))


def test_geologic_time_str_fails_when_not_int():
    with pytest.raises(AssertionError):
        rqts.geologic_time_str('hello world')


@pytest.mark.parametrize('years_value, expected_result', [(10_000_000, '-10 Ma'), (1_000_000, '-1 Ma'),
                                                          (11_000_000, '-11 Ma'), (1, '-0.000 Ma'),
                                                          (1_500_000, '-1.500 Ma'), (15_060_000, '-16 Ma'),
                                                          (15_040_000, '-16 Ma')])
def test_geologic_time_str(years_value, expected_result):
    # arrange
    test_years_value = years_value
    # act
    result = rqts.geologic_time_str(test_years_value)
    # assert
    assert result == expected_result



import pytest

import resqpy.organize as rqo
import resqpy.organize.organization_feature as OrganizationFeature
from resqpy.model import Model


# Test saving and loading from disk
@pytest.mark.parametrize(
    "cls, data",
    [
        (
            rqo.OrganizationFeature,
            dict(feature_name = 'hello', organization_kind = 'stratigraphic'),
        ),
        (
            rqo.GeobodyFeature,
            dict(feature_name = 'hi'),
        ),
        (
            rqo.BoundaryFeature,
            dict(feature_name = 'foobar'),
        ),
        (
            rqo.FrontierFeature,
            dict(feature_name = 'foobar'),
        ),
        (
            rqo.GeologicUnitFeature,
            dict(feature_name = 'foobar'),
        ),
        (
            rqo.FluidBoundaryFeature,
            dict(feature_name = 'foobar', kind = 'gas oil contact'),
        ),
        (
            rqo.TectonicBoundaryFeature,
            dict(feature_name = 'foobar', kind = 'fracture'),
        ),
        (
            rqo.GeneticBoundaryFeature,
            dict(feature_name = 'foobar', kind = 'geobody boundary'),  # Age as well?
        ),
        (
            rqo.WellboreFeature,
            dict(feature_name = 'foobar'),
        ),
    ])
def test_organize_classes(tmp_model, cls, data):

    # Load example model from a fixture
    model = tmp_model
    epc = model.epc_file

    # Create the feature
    obj = cls(parent_model = model, **data)
    uuid = obj.uuid

    # Save to disk
    obj.create_xml()
    model.store_epc()
    model.h5_release()

    # Reload from disk
    del model, obj
    model2 = Model(epc_file = epc)
    obj2 = cls(parent_model = model2, uuid = uuid)

    # Check all attributes were loaded correctly
    for key, expected_value in data.items():
        assert getattr(obj2, key) == expected_value, f"Error for {key}"


def test_RockFluidUnitFeature(tmp_model):

    # Create the features
    top = rqo.BoundaryFeature(tmp_model, feature_name = 'the top')
    base = rqo.BoundaryFeature(tmp_model, feature_name = 'the base')
    rfuf_1 = rqo.RockFluidUnitFeature(parent_model = tmp_model,
                                      feature_name = 'foobar',
                                      phase = 'seal',
                                      top_boundary_feature = top,
                                      base_boundary_feature = base)
    uuid = rfuf_1.uuid

    # Save to disk
    top.create_xml()
    base.create_xml()
    rfuf_1.create_xml()
    tmp_model.store_epc()

    # Reload from disk
    rfuf_2 = rqo.RockFluidUnitFeature(parent_model = tmp_model, uuid = uuid)

    # Check properties the same
    assert rfuf_2.feature_name == 'foobar'
    assert rfuf_2.phase == 'seal'
    assert rfuf_2.top_boundary_feature.feature_name == 'the top'
    assert rfuf_2.base_boundary_feature.feature_name == 'the base'


def test_FaultInterp(tmp_model):

    title = "not my fault"
    tect_boundary = rqo.TectonicBoundaryFeature(tmp_model, kind = 'fault')
    fault_interp = rqo.FaultInterpretation(
        tmp_model,
        tectonic_boundary_feature = tect_boundary,
        title = title,
        domain = "depth",
        is_normal = True,
        maximum_throw = 3,
    )

    tect_boundary.create_xml()
    fault_interp.create_xml()

    fault_interp_2 = rqo.FaultInterpretation(tmp_model, uuid = fault_interp.uuid)
    assert fault_interp_2.title == title
    assert fault_interp_2.maximum_throw == 3


def test_EarthModel(tmp_model):

    title = 'gaia'
    org_feat = rqo.OrganizationFeature(tmp_model, feature_name = 'marie kondo', organization_kind = "earth model")
    em1 = rqo.EarthModelInterpretation(tmp_model, title = title, organization_feature = org_feat)

    org_feat.create_xml()
    em1.create_xml()

    em2 = rqo.EarthModelInterpretation(tmp_model, uuid = em1.uuid)
    assert em2.title == title


def test_Horizon(tmp_model):

    gen = rqo.GeneticBoundaryFeature(tmp_model, kind = 'horizon')
    hor = rqo.HorizonInterpretation(tmp_model,
                                    genetic_boundary_feature = gen,
                                    sequence_stratigraphy_surface = 'maximum flooding')

    gen.create_xml()
    hor.create_xml()

    hor2 = rqo.HorizonInterpretation(tmp_model, uuid = hor.uuid)
    assert hor2.sequence_stratigraphy_surface == hor.sequence_stratigraphy_surface


def test_GeobodyBoundary(tmp_model):

    gen = rqo.GeneticBoundaryFeature(tmp_model, kind = 'geobody boundary')
    gb = rqo.GeobodyBoundaryInterpretation(tmp_model, genetic_boundary_feature = gen)

    gen.create_xml()
    gb.create_xml()
    gb2 = rqo.GeobodyBoundaryInterpretation(tmp_model, uuid = gb.uuid)
    assert gb == gb2


def test_wellbore_interp_title(tmp_model):
    # Create a feature and interp objects
    feature_name = 'well A'
    well_feature = rqo.WellboreFeature(tmp_model, feature_name = feature_name)
    well_feature.create_xml()
    well_interp_1 = rqo.WellboreInterpretation(tmp_model, wellbore_feature = well_feature, is_drilled = True)
    well_interp_1.create_xml()

    # Create a duplicate object, loading from XML
    well_interp_2 = rqo.WellboreInterpretation(tmp_model, uuid = well_interp_1.uuid)

    # Check feature name is present
    assert well_interp_1.title == feature_name
    assert well_interp_2.title == feature_name



"""Base class for generic resqml objects."""

import logging
import warnings
from abc import ABCMeta, abstractmethod

import resqpy.olio.uuid as bu
import resqpy.olio.xml_et as rqet

logger = logging.getLogger(__name__)


class BaseResqpy(metaclass = ABCMeta):
    """Base class for generic resqpy classes.

    Implements generic attributes such as uuid, root, part, title, originator.

    Implements generic magic methods, such as pretty printing and testing for
    equality.

    Example use::

        class AnotherResqpyObject(BaseResqpy):

            resqml_type = 'obj_anotherresqmlobjectrepresentation'
    """

    @property
    @abstractmethod
    def resqml_type(self):
        """Definition of which RESQML object the class represents.

        Subclasses must overwrite this abstract attribute.
        """
        raise NotImplementedError

    def __init__(self, model, uuid = None, title = None, originator = None, root_node = None, extra_metadata = None):
        """Load an existing resqml object, or create new.

        Args:
            model (resqpy.model.Model): Parent model
            uuid (str, optional): Load from existing uuid (if given), else create new.
            title (str, optional): Citation title
            originator (str, optional): Creator of object. By default, uses user id.
        """
        self.model = model
        self.title = title  #: Citation title
        self.originator = originator  #: Creator of object. By default, user id.
        self.extra_metadata = {}
        if extra_metadata:
            self.extra_metadata = extra_metadata
            self._standardise_extra_metadata()  # has side effect of making a copy

        if root_node is not None:
            warnings.warn("root_node parameter is deprecated, use uuid instead", DeprecationWarning)
            uuid = rqet.uuid_for_part_root(root_node)

        if uuid is None:
            self.uuid = bu.new_uuid()  #: Unique identifier


#            logger.debug(f"created new uuid for object {self}")
        else:
            self.uuid = uuid
            #            logger.debug(f"loading existing object {self}")
            citation_node = rqet.find_tag(self.root, 'Citation')
            if citation_node is not None:
                self.title = rqet.find_tag_text(citation_node, 'Title')
                self.originator = rqet.find_tag_text(citation_node, 'Originator')
            self.extra_metadata = rqet.load_metadata_from_xml(self.root)
            self._load_from_xml()

    # usually overridden by derived class, unless generic code above handles all attributes
    def _load_from_xml(self):
        pass

    # Define attributes self.part and self.root, using uuid as the primary key

    @property
    def part(self):
        """Standard part name corresponding to self.uuid."""

        # following caused trouble when resqml_type dynamically determined
        #       return rqet.part_name_for_object(self.resqml_type, self.uuid)
        return self.model.part_for_uuid(self.uuid)

    @property
    def root(self):
        """XML node corresponding to self.uuid."""

        return self.model.root_for_uuid(self.uuid)

    @property
    def citation_title(self):
        """Citation block title equivalent to self.title."""

        return self.title

    def try_reuse(self):
        """Look for an equivalent existing RESQML object and modify the uuid of this object if found.

        returns:
           boolean: True if an equivalent object was found, False if not

        note:
           by design this method may change this object's uuid as a side effect
        """

        assert self.uuid is not None
        if self.root is not None:
            return True
        uuid_list = self.model.uuids(obj_type = self.resqml_type)
        for other_uuid in uuid_list:
            if bu.matching_uuids(self.uuid, other_uuid):
                logger.debug(f'reusing existing xml for uuid {other_uuid}')
                return True
            try:
                other = self.__class__(self.model, uuid = other_uuid)
            except Exception:
                return False
            if self == other:
                logger.debug(f'reusing equivalent resqml object with uuid {other_uuid}')
                self.uuid = other_uuid  # NB: change of uuid for this object
                assert self.root is not None
                return True
        return False

    def create_xml(self, title = None, originator = None, extra_metadata = None, add_as_part = False):
        """Write citation block to XML.

        Note:

            `add_as_part` is False by default in this base method. Derived classes should typically
            extend this method to complete the XML representation, and then finally ensure the node
            is added as a part to the model.

        Args:
            title (string): used as the citation Title text
            originator (string, optional): the name of the human being who created the deviation survey part;
                default is to use the login name
            extra_metadata (dict, optional): extra metadata items to be added
            add_as_part (boolean): if True, the newly created xml node is added as a part
                in the model

        Returns:
            node: the newly created root node
        """

        assert self.uuid is not None

        # Create the root node
        node = self.model.new_obj_node(self.resqml_type)
        node.attrib['uuid'] = str(self.uuid)

        # Citation block
        if title:
            self.title = title
        if originator:
            self.originator = originator
        self.model.create_citation(root = node, title = self.title, originator = self.originator)

        # Extra metadata
        if extra_metadata:
            if not hasattr(self, 'extra_metadata'):
                self.extra_metadata = {}
            for key, value in extra_metadata.items():
                self.extra_metadata[str(key)] = str(value)
        if hasattr(self, 'extra_metadata') and self.extra_metadata:
            rqet.create_metadata_xml(node = node, extra_metadata = self.extra_metadata)

        if add_as_part:
            self.model.add_part(self.resqml_type, self.uuid, node)
            assert self.root is not None

        return node

    def append_extra_metadata(self, meta_dict):
        """Append a given dictionary of metadata to the existing metadata."""
        for key in meta_dict:
            self.extra_metadata[key] = meta_dict[key]
        self._standardise_extra_metadata()

    def _standardise_extra_metadata(self):
        if self.extra_metadata:
            em = {}
            for key, value in self.extra_metadata.items():
                em[str(key)] = str(value)
            self.extra_metadata = em

    # Generic magic methods

    def __eq__(self, other):
        """Implements equals operator; uses is_equivalent() otherwise compares class type and uuid."""
        if hasattr(self, 'is_equivalent'):
            return self.is_equivalent(other)
        if not isinstance(other, self.__class__):
            return False
        other_uuid = getattr(other, "uuid", None)
        return bu.matching_uuids(self.uuid, other_uuid)

    def __ne__(self, other):
        """Implements not equal operator."""
        return not self.__eq__(other)

    def __repr__(self):
        """String representation."""
        return f"{self.__class__.__name__}(uuid={self.uuid}, title={self.title})"

    def _repr_html_(self):
        """Return HTML for IPython / Jupyter representation."""

        keys_to_display = ('uuid', 'title', 'originator')
        html = f"<h3>{self.__class__.__name__}</h3>\n"
        for key in keys_to_display:
            html += f"<strong>{key}</strong>: {getattr(self, key)}<br>\n"
        return html

    # Include some aliases for root, but raise warnings if they are used
    # TODO: remove these aliases for self.root

    @property
    def root_node(self):
        """DEPRECATED.

        Alias for root
        """
        warnings.warn("Attribute 'root_node' is deprecated. Use 'root'", DeprecationWarning)
        return self.root

    @property
    def node(self):
        """DEPRECATED.

        Alias for root
        """
        warnings.warn("Attribute 'node' is deprecated. Use 'root'", DeprecationWarning)
        return self.root



import os

import numpy as np
import pandas as pd
import pytest
from numpy.testing import assert_array_almost_equal

import resqpy.fault as rqf
import resqpy.grid as grr
import resqpy.model as rq
import resqpy.olio.xml_et as rqet
import resqpy.property as rqp


# yapf: disable
@pytest.mark.parametrize('inc_list,tmult_dict,expected_mult',
   [(['fault_1.inc'], {}, {'fault_1': 1}),
    (['fault_1.inc'], {'fault_1': 2}, {'fault_1': 2}),
    (['fault_1.inc', 'fault_2.inc'], {'fault_1': 2}, {'fault_1': 2, 'fault_2': 2})])
# yapf: enable
def test_add_connection_set_and_tmults(example_model_with_properties, test_data_path, inc_list, tmult_dict,
                                       expected_mult):
    model = example_model_with_properties

    inc_list = [os.path.join(test_data_path, inc) for inc in inc_list]

    gcs_uuid = rqf.add_connection_set_and_tmults(model, inc_list, tmult_dict)

    assert gcs_uuid is not None, 'Grid connection set not generated'

    reload_model = rq.Model(epc_file = model.epc_file)

    faults = reload_model.parts_list_of_type(('obj_FaultInterpretation'))
    assert len(faults) == len(
        expected_mult.keys()), f'Expected a {len(expected_mult.keys())} faults, found {len(faults)}'
    for fault in faults:
        metadata = rqet.load_metadata_from_xml(reload_model.root_for_part(fault))
        title = reload_model.citation_title_for_part(fault)
        expected_str = str(float(expected_mult[title]))
        assert metadata["Transmissibility multiplier"] == expected_str,  \
            f'Expected mult for fault {title} to be {expected_str}, found {metadata["Transmissibility multiplier"]}'

    # check that a transmissibility multiplier property has been created
    gcs = rqf.GridConnectionSet(reload_model, uuid = gcs_uuid, find_properties = True)
    assert gcs is not None
    pc = gcs.property_collection
    assert pc is not None and pc.number_of_parts() > 0
    part = pc.singleton(property_kind = 'transmissibility multiplier')
    assert part is not None
    # check property values are in expected set
    a = pc.cached_part_array_ref(part)
    assert a is not None and a.ndim == 1
    expect = [x for x in expected_mult.values()]
    assert all([v in expect for v in a])
    # see if a local property kind has been set up correctly
    pku = pc.local_property_kind_uuid(part)
    assert pku is not None
    pk = rqp.PropertyKind(reload_model, uuid = pku)
    assert pk is not None
    assert pk.title == 'transmissibility multiplier'


def test_gcs_property_inheritance(tmp_path):

    epc = os.path.join(tmp_path, 'gcs_prop_inherit.epc')

    model = rq.Model(epc, new_epc = True, create_basics = True, create_hdf5_ext = True)

    # create a grid
    g = grr.RegularGrid(model, (5, 3, 3), dxyz = (10.0, 10.0, 1.0))
    g.write_hdf5()
    g.create_xml(title = 'unsplit grid')

    # define an L shaped (in plan view) fault
    j_faces = np.zeros((g.nk, g.nj - 1, g.ni), dtype = bool)
    j_faces[:, 0, 1:] = True
    i_faces = np.zeros((g.nk, g.nj, g.ni - 1), dtype = bool)
    i_faces[:, 1:, 0] = True
    gcs = rqf.GridConnectionSet(model,
                                grid = g,
                                j_faces = j_faces,
                                i_faces = i_faces,
                                feature_name = 'L fault',
                                create_organizing_objects_where_needed = True,
                                create_transmissibility_multiplier_property = False)
    # check that connection set has the right number of cell face pairs
    assert gcs.count == g.nk * ((g.nj - 1) + (g.ni - 1))

    # create a transmissibility multiplier property
    tm = np.arange(gcs.count).astype(float)
    if gcs.property_collection is None:
        gcs.property_collection = rqp.PropertyCollection()
        gcs.property_collection.set_support(support = gcs)
    pc = gcs.property_collection
    pc.add_cached_array_to_imported_list(
        tm,
        'unit test',
        'TMULT',
        uom = 'Euc',  # actually a ratio of transmissibilities
        property_kind = 'transmissibility multiplier',
        local_property_kind_uuid = None,
        realization = None,
        indexable_element = 'faces')
    # write gcs which should also write property collection and create a local property kind
    gcs.write_hdf5()
    gcs.create_xml(write_new_properties = True)

    # check that a local property kind has materialised
    pk_uuid = model.uuid(obj_type = 'PropertyKind', title = 'transmissibility multiplier')
    assert pk_uuid is not None

    # create a derived grid connection set using a layer range
    thin_gcs, thin_indices = gcs.filtered_by_layer_range(min_k0 = 1, max_k0 = 3, return_indices = True)
    assert thin_gcs is not None and thin_indices is not None
    assert thin_gcs.count == 3 * ((g.nj - 1) + (g.ni - 1))
    # inherit the transmissibility multiplier property
    thin_gcs.inherit_properties_for_selected_indices(gcs, thin_indices)
    thin_gcs.write_hdf5()
    thin_gcs.create_xml()  # by default will include write of new properties

    # check that the inheritance has worked
    assert thin_gcs.property_collection is not None and thin_gcs.property_collection.number_of_parts() > 0
    thin_pc = thin_gcs.property_collection
    tm_part = thin_pc.singleton(property_kind = 'transmissibility multiplier')
    assert tm_part is not None
    thin_tm = thin_pc.cached_part_array_ref(tm_part)
    assert thin_tm is not None and thin_tm.ndim == 1
    assert thin_tm.size == thin_gcs.count
    assert_array_almost_equal(thin_tm, tm[thin_indices])

    # check that get_combined...() method can execute using property collection
    b_a, i_a, f_a = gcs.get_combined_fault_mask_index_value_arrays(min_k = 1,
                                                                   max_k = 3,
                                                                   property_name = 'Transmissibility multiplier',
                                                                   ref_k = 2)
    assert b_a is not None and i_a is not None and f_a is not None
    # check that transmissibility multiplier values have been sampled correctly from property array
    assert f_a.shape == (g.nj, g.ni, 2, 2)
    assert np.count_nonzero(np.isnan(f_a)) == 4 * g.nj * g.ni - 2 * ((g.nj - 1) + (g.ni - 1))
    assert np.nanmax(f_a) > np.nanmin(f_a)
    restore = np.seterr(all = 'ignore')
    assert np.all(np.logical_or(np.isnan(f_a), f_a >= np.nanmin(thin_tm)))
    assert np.all(np.logical_or(np.isnan(f_a), f_a <= np.nanmax(thin_tm)))
    np.seterr(**restore)


def test_pinchout_and_k_gap_gcs(tmp_path):

    epc = os.path.join(tmp_path, 'gcs_pinchout_k_gap.epc')
    model = rq.new_model(epc)

    # create a grid
    g = grr.RegularGrid(model, (5, 5, 5), dxyz = (100.0, 100.0, 10.0), as_irregular_grid = True)
    # patch points to generate a pinchout
    p = g.points_cached
    assert p.shape == (6, 6, 6, 3)
    p[2, :3, :3] = p[1, :3, :3]
    # convert one layer to a K gap with pinchout
    p[4, 3:, 3:] = p[3, 3:, 3:]
    g.nk -= 1
    g.extent_kji = np.array((g.nk, g.nj, g.ni), dtype = int)
    g.k_gaps = 1
    g.k_gap_after_array = np.zeros(g.nk - 1, dtype = bool)
    g.k_gap_after_array[2] = True
    g._set_k_raw_index_array()
    g.write_hdf5()
    g.create_xml(title = 'pinchout k gap grid')
    model.store_epc()

    # reload the grid
    model = rq.Model(epc)
    grid = model.grid()
    assert grid is not None
    assert grid.k_gaps == 1
    assert tuple(grid.extent_kji) == (4, 5, 5)

    # create a pinchout connection set
    po_gcs = rqf.pinchout_connection_set(grid)
    assert po_gcs is not None
    po_gcs.write_hdf5()
    po_gcs.create_xml()
    po_uuid = po_gcs.uuid

    # create a K gap connection set
    kg_gcs = rqf.k_gap_connection_set(grid)
    assert kg_gcs is not None
    kg_gcs.write_hdf5()
    kg_gcs.create_xml()
    kg_uuid = kg_gcs.uuid

    model.store_epc()

    # re-open the model and load the connection sets
    model = rq.Model(epc)
    po_gcs = rqf.GridConnectionSet(model, uuid = po_uuid)
    assert po_gcs is not None
    po_gcs.cache_arrays()
    kg_gcs = rqf.GridConnectionSet(model, uuid = kg_uuid)
    assert kg_gcs is not None
    kg_gcs.cache_arrays()

    # check face pairs in the pinchout connection set
    assert po_gcs.count == 4
    assert po_gcs.cell_index_pairs.shape == (4, 2)
    assert po_gcs.face_index_pairs.shape == (4, 2)
    assert np.all(po_gcs.cell_index_pairs[:, 0] != po_gcs.cell_index_pairs[:, 1])
    assert np.all(po_gcs.face_index_pairs[:, 0] != po_gcs.cell_index_pairs[:, 1])
    assert np.all(np.logical_or(po_gcs.face_index_pairs == 0, po_gcs.face_index_pairs == 1))
    for cell in po_gcs.cell_index_pairs.flatten():
        assert cell in [0, 1, 5, 6, 50, 51, 55, 56]
    assert np.all(np.abs(po_gcs.cell_index_pairs[:, 1] - po_gcs.cell_index_pairs[:, 0]) == 50)

    # check face pairs in K gap connection set
    assert kg_gcs.count == 4
    assert kg_gcs.cell_index_pairs.shape == (4, 2)
    assert kg_gcs.face_index_pairs.shape == (4, 2)
    assert np.all(kg_gcs.cell_index_pairs[:, 0] != kg_gcs.cell_index_pairs[:, 1])
    assert np.all(kg_gcs.face_index_pairs[:, 0] != kg_gcs.cell_index_pairs[:, 1])
    assert np.all(np.logical_or(kg_gcs.face_index_pairs == 0, kg_gcs.face_index_pairs == 1))
    for cell in kg_gcs.cell_index_pairs.flatten():
        assert cell in [74, 73, 69, 68, 99, 98, 94, 93]
    assert np.all(np.abs(kg_gcs.cell_index_pairs[:, 1] - kg_gcs.cell_index_pairs[:, 0]) == 25)

    # test compact indices method
    ci = po_gcs.compact_indices()
    assert ci.shape == (4, 2)
    for cf in ci.flatten():
        assert cf in [1, 7, 31, 37, 300, 306, 330, 336]
    assert np.all(np.abs(ci[:, 1] - ci[:, 0]) == 299)

    # test write simulator method
    files = ('fault_ff.dat', 'fault_tf.dat', 'fault_ft.dat')
    both_sides = (False, True, False)
    minus = (False, False, True)
    for filename, inc_both_sides, use_minus in zip(files, both_sides, minus):
        dat_path = os.path.join(tmp_path, filename)
        po_gcs.write_simulator(dat_path, include_both_sides = inc_both_sides, use_minus = use_minus)
        assert os.path.exists(dat_path)


def test_two_fault_gcs(tmp_path):

    epc = os.path.join(tmp_path, 'two_fault.epc')
    model = rq.new_model(epc)

    # create a grid
    g = grr.RegularGrid(model, (5, 4, 3), dxyz = (100.0, 100.0, 10.0))
    g.create_xml()

    # create an empty grid connection set
    gcs = rqf.GridConnectionSet(model, grid = g)

    # prepare two named faults as a dataframe
    df = pd.DataFrame(columns = ['name', 'face', 'i1', 'i2', 'j1', 'j2', 'k1', 'k2', 'mult'])
    df = df.append({
        'name': 'F1',
        'face': 'I+',
        'i1': 1,
        'i2': 1,
        'j1': 0,
        'j2': 3,
        'k1': 0,
        'k2': 4,
        'mult': 0.1
    },
                   ignore_index = True)
    df = df.append({
        'name': 'F2',
        'face': 'J-',
        'i1': 0,
        'i2': 2,
        'j1': 2,
        'j2': 2,
        'k1': 0,
        'k2': 4,
        'mult': 0.05
    },
                   ignore_index = True)

    # set grid connection set from dataframe
    gcs.set_pairs_from_faces_df(df,
                                create_organizing_objects_where_needed = True,
                                create_mult_prop = True,
                                fault_tmult_dict = None,
                                one_based_indexing = False)

    # save the grid connection set
    gcs.write_hdf5()
    gcs.create_xml(title = 'two fault gcs')
    gcs_uuid = gcs.uuid
    model.store_epc()

    # re-open the model and check the gcs
    model = rq.Model(epc)
    gcs = rqf.GridConnectionSet(model, uuid = gcs_uuid)
    assert gcs is not None
    assert gcs.number_of_features() == 2
    feature_names = gcs.list_of_feature_names()
    assert len(feature_names) == 2
    assert 'F1' in feature_names and 'F2' in feature_names
    fault_names = gcs.list_of_fault_names()
    assert fault_names == feature_names
    for fi in (0, 1):
        assert gcs.feature_name_for_feature_index(fi) in ('F1', 'F2')
    assert gcs.feature_name_for_feature_index(0) != gcs.feature_name_for_feature_index(1)
    fi, f_uuid = gcs.feature_index_and_uuid_for_fault_name('F1')
    assert fi is not None and fi in (0, 1)
    assert f_uuid is not None
    assert gcs.fault_name_for_feature_index(fi) == 'F1'
    assert gcs.feature_index_for_cell_face((1, 1, 1), 0, 1) is None
    fi_a = gcs.feature_index_for_cell_face((1, 1, 1), 2, 1)
    assert fi_a in (0, 1)
    fi_b = gcs.feature_index_for_cell_face((1, 1, 1), 1, 1)
    assert fi_b in (0, 1)
    assert fi_a != fi_b
    gcs.rework_face_pairs()



# tests for resqpy.olio.fine_coarse class and functions

import numpy as np
import pytest
from numpy.testing import assert_array_almost_equal

import resqpy.olio.fine_coarse as rqfc


def test_fine_coarse():

    fc1 = rqfc.FineCoarse((3, 4, 5), (3, 4, 5))
    assert fc1 is not None
    fc1.set_all_ratios_constant()
    fc1.assert_valid()
    assert fc1.ratios((2, 2, 2)) == (1, 1, 1)

    fc2 = rqfc.FineCoarse((9, 16, 10), (3, 4, 5))
    assert fc2 is not None
    fc2.set_all_ratios_constant()
    fc2.set_all_proprtions_equal()
    fc2.assert_valid()
    assert fc2.ratios((2, 2, 2)) == (3, 4, 2)

    assert fc2.coarse_for_fine_kji0((5, 5, 5)) == (1, 1, 2)
    assert fc2.coarse_for_fine_axial(1, 13) == 3
    assert fc2.fine_base_for_coarse_axial(0, 1) == 3
    assert fc2.fine_base_for_coarse((2, 2, 2)) == (6, 8, 4)
    assert np.all(fc2.fine_box_for_coarse((1, 2, 3)) == np.array([[3, 8, 6], [5, 11, 7]], dtype = int))

    assert_array_almost_equal(fc2.proportion(1, 1), (0.25, 0.25, 0.25, 0.25))
    assert_array_almost_equal(fc2.interpolation(1, 1), (0.0, 0.25, 0.5, 0.75))

    # todo: several more methods to test


def test_fine_coarse_functions():

    assert rqfc.axis_for_letter('k') == 0
    assert rqfc.axis_for_letter('J') == 1
    assert rqfc.axis_for_letter('i') == 2
    assert rqfc.letter_for_axis(0) == 'K'
    assert rqfc.letter_for_axis(2) == 'I'

    fct = rqfc.tartan_refinement(coarse_extent_kji = (5, 7, 7),
                                 coarse_fovea_box = np.array([[0, 2, 2], [1, 4, 4]], dtype = int),
                                 fovea_ratios_kji = (3, 3, 4),
                                 decay_rates_kji = (1, 1, 1),
                                 decay_mode = 'linear')
    assert fct is not None
    fct.assert_valid()

    assert fct.coarse_extent_kji == (5, 7, 7)
    assert fct.fine_extent_kji == (10, 15, 22)
    assert np.all(fct.coarse_for_fine_axial_vector(0) == (0, 0, 0, 1, 1, 1, 2, 2, 3, 4))
    for const in fct.constant_ratios:
        assert const is None
    assert np.all(fct.vector_ratios[1] == (1, 2, 3, 3, 3, 2, 1))
    assert np.all(fct.vector_ratios[2] == (2, 3, 4, 4, 4, 3, 2))

    # todo: add test of tartan refinement in exponential mode



"""Class for RESQML Tectonic Boundary Feature (fault) organizational objects."""

from ._utils import (equivalent_extra_metadata, alias_for_attribute, extract_has_occurred_during,
                     equivalent_chrono_pairs, create_xml_has_occurred_during)

import resqpy.olio.uuid as bu
import resqpy.olio.xml_et as rqet
from resqpy.olio.base import BaseResqpy
from resqpy.olio.xml_namespaces import curly_namespace as ns


class TectonicBoundaryFeature(BaseResqpy):
    """Class for RESQML Tectonic Boundary Feature (fault) organizational objects."""

    resqml_type = "TectonicBoundaryFeature"
    feature_name = alias_for_attribute("title")
    valid_kinds = ('fault', 'fracture')

    def __init__(self,
                 parent_model,
                 root_node = None,
                 uuid = None,
                 kind = None,
                 feature_name = None,
                 extra_metadata = None):
        """Initialises a tectonic boundary feature (fault or fracture) organisational object."""
        self.kind = kind
        super().__init__(model = parent_model,
                         uuid = uuid,
                         title = feature_name,
                         extra_metadata = extra_metadata,
                         root_node = root_node)

    def _load_from_xml(self):
        self.kind = rqet.find_tag_text(self.root, 'TectonicBoundaryKind')
        if not self.kind:
            self.kind = 'fault'

    def is_equivalent(self, other, check_extra_metadata = True):
        """Returns True if this feature is essentially the same as the other; otherwise False."""
        if other is None or not isinstance(other, TectonicBoundaryFeature):
            return False
        if self is other or bu.matching_uuids(self.uuid, other.uuid):
            return True
        if check_extra_metadata and not equivalent_extra_metadata(self, other):
            return False
        return self.feature_name == other.feature_name and self.kind == other.kind

    def create_xml(self, add_as_part = True, originator = None, reuse = True):
        """Creates a tectonic boundary feature organisational xml node from this tectonic boundary feature object."""

        if reuse and self.try_reuse():
            return self.root  # check for reusable (equivalent) object
        # create node with citation block
        tbf = super().create_xml(add_as_part = False, originator = originator)

        assert self.kind in self.valid_kinds
        kind_node = rqet.SubElement(tbf, ns['resqml2'] + 'TectonicBoundaryKind')
        kind_node.set(ns['xsi'] + 'type', ns['resqml2'] + 'TectonicBoundaryKind')
        kind_node.text = self.kind

        if add_as_part:
            self.model.add_part('obj_TectonicBoundaryFeature', self.uuid, tbf)

        return tbf



"""Miscellaneous functions for importing from other formats."""

# Nexus is a trademark of Halliburton

__all__ = [
    'import_nexus', 'import_vdb_all_grids', 'grid_from_cp', 'import_vdb_ensemble', 'add_ab_properties', 'add_surfaces'
]

from ._grid_from_cp import grid_from_cp
from ._import_nexus import import_nexus
from ._import_vdb_all_grids import import_vdb_all_grids
from ._import_vdb_ensemble import import_vdb_ensemble
from ._add_ab_properties import add_ab_properties
from ._add_surfaces import add_surfaces



"""Submodule containing functions for adding properties to a property collection."""

version = '1st December 2021'

# Nexus is a registered trademark of the Halliburton Company

import logging

log = logging.getLogger(__name__)

import resqpy.olio.uuid as bu
import resqpy.olio.xml_et as rqet

from .property_common import same_property_kind, guess_uom, property_kind_and_facet_from_keyword
import resqpy.property._collection_get_attributes as pcga


def _add_selected_part_from_other_dict(collection, part, other, realization, support_uuid, uuid, continuous,
                                       categorical, count, points, indexable, property_kind, facet_type, facet,
                                       citation_title, citation_title_match_starts_with, time_series_uuid, time_index,
                                       string_lookup_uuid, ignore_clashes):
    if _check_not_none_and_not_equals(realization, other.realization_for_part, part):
        return
    if _check_not_none_and_not_uuid_match(support_uuid, other.support_uuid_for_part, part):
        return
    if _check_not_none_and_not_uuid_match(uuid, other.uuid_for_part, part):
        return
    if _check_not_none_and_not_equals(continuous, other.continuous_for_part, part):
        return
    if _check_categorical_and_lookup(categorical, other, part):
        return
    if _check_not_none_and_not_equals(count, other.count_for_part, part):
        return
    if _check_not_none_and_not_equals(points, other.points_for_part, part):
        return
    if _check_not_none_and_not_equals(indexable, other.indexable_for_part, part):
        return
    if property_kind is not None and not same_property_kind(other.property_kind_for_part(part), property_kind):
        return
    if _check_not_none_and_not_equals(facet_type, other.facet_type_for_part, part):
        return
    if _check_not_none_and_not_equals(facet, other.facet_for_part, part):
        return
    if _check_citation_title(citation_title, citation_title_match_starts_with, other, part):
        return
    if _check_not_none_and_not_uuid_match(time_series_uuid, other.time_series_uuid_for_part, part):
        return
    if _check_not_none_and_not_equals(time_index, other.time_index_for_part, part):
        return
    if _check_not_none_and_not_uuid_match(string_lookup_uuid, other.string_lookup_uuid_for_part, part):
        return
    if part in collection.dict.keys():
        if ignore_clashes:
            return
        assert (False)
    collection.dict[part] = other.dict[part]


def _add_part_to_dict_get_realization(collection, realization, xml_node):
    if realization is not None and collection.realization is not None:
        assert (realization == collection.realization)
    if realization is None:
        realization = collection.realization
    realization_node = rqet.find_tag(xml_node, 'RealizationIndex')  # optional; if present use to populate realization
    if realization_node is not None:
        realization = int(realization_node.text)
    return realization


def _add_part_to_dict_get_type_details(collection, part, continuous, xml_node):
    sl_ref_node = None
    type = collection.model.type_of_part(part)
    #      log.debug('adding part ' + part + ' of type ' + type)
    assert type in ['obj_ContinuousProperty', 'obj_DiscreteProperty', 'obj_CategoricalProperty', 'obj_PointsProperty']
    if continuous is None:
        continuous = (type in ['obj_ContinuousProperty', 'obj_PointsProperty'])
    else:
        assert continuous == (type in ['obj_ContinuousProperty', 'obj_PointsProperty'])
    points = (type == 'obj_PointsProperty')
    string_lookup_uuid = None
    if type == 'obj_CategoricalProperty':
        sl_ref_node = rqet.find_tag(xml_node, 'Lookup')
        string_lookup_uuid = bu.uuid_from_string(rqet.find_tag_text(sl_ref_node, 'UUID'))

    return type, continuous, points, string_lookup_uuid, sl_ref_node


def _add_part_to_dict_get_support_uuid(collection, part):
    support_uuid = collection.model.supporting_representation_for_part(part)
    if support_uuid is None:
        support_uuid = collection.support_uuid
    elif collection.support_uuid is None:
        collection.set_support(support_uuid)
    elif not bu.matching_uuids(support_uuid, collection.support.uuid):  # multi-support collection
        collection.set_support(None)
    if isinstance(support_uuid, str):
        support_uuid = bu.uuid_from_string(support_uuid)
    return support_uuid


def _add_part_to_dict_get_uom(collection, part, continuous, xml_node, trust_uom, property_kind, minimum, maximum, facet,
                              facet_type):
    uom = None
    if continuous:
        uom_node = rqet.find_tag(xml_node, 'UOM')
        if uom_node is not None and (trust_uom or uom_node.text not in ['', 'Euc']):
            uom = uom_node.text
        else:
            if not collection.guess_warning:
                collection.guess_warning = True
                log.warning("Guessing unit of measure for one or more properties")
            uom = guess_uom(property_kind, minimum, maximum, collection.support, facet_type = facet_type, facet = facet)
    return uom


def _check_not_none_and_not_equals(attrib, method, part):
    return attrib is not None and method(part) != attrib


def _check_not_none_and_not_uuid_match(uuid, method, part):
    return uuid is not None and not bu.matching_uuids(uuid, method(part))


def _check_citation_title(citation_title, citation_title_match_starts_with, other, part):
    if citation_title is not None:
        if citation_title_match_starts_with:
            if not other.citation_title_for_part(part).startswith():
                return True
        else:
            if other.citation_title_for_part(part) != citation_title:
                return True
    return False


def _check_categorical_and_lookup(categorical, other, part):
    if categorical is not None:
        if categorical:
            if other.continuous_for_part(part) or (other.string_lookup_uuid_for_part(part) is None):
                return True
        else:
            if not (other.continuous_for_part(part) or (other.string_lookup_uuid_for_part(part) is None)):
                return True
    return False


def _process_imported_property(collection, attributes, property_kind_uuid, string_lookup_uuid, time_series_uuid,
                               ext_uuid, support_uuid, selected_time_indices_list, find_local_property_kinds,
                               extra_metadata, expand_const_arrays):
    (p_uuid, p_file_name, p_keyword, p_cached_name, p_discrete, p_uom, p_time_index, p_null_value, p_min_value,
     p_max_value, property_kind, facet_type, facet, realization, indexable_element, count, local_property_kind_uuid,
     const_value, points) = attributes

    log.debug('processing imported property ' + str(p_keyword))
    assert not points or not p_discrete
    if local_property_kind_uuid is None:
        local_property_kind_uuid = property_kind_uuid

    property_kind = _process_imported_property_get_property_kind(collection, property_kind, local_property_kind_uuid,
                                                                 p_keyword, p_discrete, string_lookup_uuid, points)

    p_array = _process_imported_property_get_p_array(collection, p_cached_name)

    add_min_max = pcga._process_imported_property_get_add_min_max(points, property_kind, string_lookup_uuid,
                                                                  local_property_kind_uuid)

    if selected_time_indices_list is not None and p_time_index is not None:
        p_time_index = selected_time_indices_list.index(p_time_index)
    p_node = collection.create_xml(
        ext_uuid = ext_uuid,
        property_array = p_array,
        title = p_keyword,
        property_kind = property_kind,
        support_uuid = support_uuid,
        p_uuid = p_uuid,
        facet_type = facet_type,
        facet = facet,
        discrete = p_discrete,  # todo: time series bits
        time_series_uuid = time_series_uuid,
        time_index = p_time_index,
        uom = p_uom,
        null_value = p_null_value,
        originator = None,
        source = p_file_name,
        add_as_part = True,
        add_relationships = True,
        add_min_max = add_min_max,
        min_value = p_min_value,
        max_value = p_max_value,
        realization = realization,
        string_lookup_uuid = string_lookup_uuid,
        property_kind_uuid = local_property_kind_uuid,
        indexable_element = indexable_element,
        count = count,
        points = points,
        find_local_property_kinds = find_local_property_kinds,
        extra_metadata = extra_metadata,
        const_value = const_value,
        expand_const_arrays = expand_const_arrays)
    if p_node is not None:
        return p_node
    else:
        return None


def _process_imported_property_get_property_kind(collection, property_kind, local_property_kind_uuid, p_keyword,
                                                 p_discrete, string_lookup_uuid, points):
    if property_kind is None:
        if local_property_kind_uuid is not None:
            # note: requires local property kind to be present
            property_kind = collection.model.title(uuid = local_property_kind_uuid)
        else:
            # todo: only if None in ab_property_list
            (property_kind, facet_type, facet) = property_kind_and_facet_from_keyword(p_keyword)
    if property_kind is None:
        # todo: the following are abstract standard property kinds, which shouldn't really have data directly associated with them
        if p_discrete:
            if string_lookup_uuid is not None:
                property_kind = 'categorical'
            else:
                property_kind = 'discrete'
        elif points:
            property_kind = 'length'
        else:
            property_kind = 'continuous'
    return property_kind


def _process_imported_property_get_p_array(collection, p_cached_name):
    if hasattr(collection, p_cached_name):
        return collection.__dict__[p_cached_name]
    else:
        return None


def _add_part_to_dict_get_count_and_indexable(xml_node):
    count_node = rqet.find_tag(xml_node, 'Count')
    assert count_node is not None
    count = int(count_node.text)

    indexable_node = rqet.find_tag(xml_node, 'IndexableElement')
    assert indexable_node is not None
    indexable = indexable_node.text

    return count, indexable


def _add_part_to_dict_get_property_kind(xml_node, citation_title):
    (p_kind_from_keyword, facet_type, facet) = property_kind_and_facet_from_keyword(citation_title)
    prop_kind_node = rqet.find_tag(xml_node, 'PropertyKind')
    assert (prop_kind_node is not None)
    kind_node = rqet.find_tag(prop_kind_node, 'Kind')
    property_kind_uuid = None  # only used for bespoke (local) property kinds
    if kind_node is not None:
        property_kind = kind_node.text  # could check for consistency with that derived from citation title
        lpk_node = None
    else:
        lpk_node = rqet.find_tag(prop_kind_node, 'LocalPropertyKind')
        if lpk_node is not None:
            property_kind = rqet.find_tag_text(lpk_node, 'Title')
            property_kind_uuid = rqet.find_tag_text(lpk_node, 'UUID')
    assert property_kind is not None and len(property_kind) > 0
    if (p_kind_from_keyword and p_kind_from_keyword != property_kind and
        (p_kind_from_keyword not in ['cell length', 'length', 'thickness'] or
         property_kind not in ['cell length', 'length', 'thickness'])):
        log.warning(
            f'property kind {property_kind} not the expected {p_kind_from_keyword} for keyword {citation_title}')
    return property_kind, property_kind_uuid, lpk_node


def _add_part_to_dict_get_facet(xml_node):
    facet_type = None
    facet = None
    facet_node = rqet.find_tag(xml_node, 'Facet')  # todo: handle more than one facet for a property
    if facet_node is not None:
        facet_type = rqet.find_tag(facet_node, 'Facet').text
        facet = rqet.find_tag(facet_node, 'Value').text
        if facet_type is not None and facet_type == '':
            facet_type = None
        if facet is not None and facet == '':
            facet = None
    return facet_type, facet


def _add_part_to_dict_get_timeseries(xml_node):
    time_series_uuid = None
    time_index = None
    time_node = rqet.find_tag(xml_node, 'TimeIndex')
    if time_node is not None:
        time_index = int(rqet.find_tag(time_node, 'Index').text)
        time_series_uuid = bu.uuid_from_string(rqet.find_tag(rqet.find_tag(time_node, 'TimeSeries'), 'UUID').text)

    return time_series_uuid, time_index


def _add_part_to_dict_get_minmax(xml_node):
    minimum = None
    min_node = rqet.find_tag(xml_node, 'MinimumValue')
    if min_node is not None:
        minimum = min_node.text  # NB: left as text
    maximum = None
    max_node = rqet.find_tag(xml_node, 'MaximumValue')
    if max_node is not None:
        maximum = max_node.text  # NB: left as text

    return minimum, maximum


def _add_part_to_dict_get_null_constvalue_points(xml_node, continuous, points):
    null_value = None
    if not continuous:
        null_value = rqet.find_nested_tags_int(xml_node, ['PatchOfValues', 'Values', 'NullValue'])
    const_value = None
    if points:
        values_node = rqet.find_nested_tags(xml_node, ['PatchOfPoints', 'Points'])
    else:
        values_node = rqet.find_nested_tags(xml_node, ['PatchOfValues', 'Values'])
    values_type = rqet.node_type(values_node)
    assert values_type is not None
    if values_type.endswith('ConstantArray'):
        if continuous:
            const_value = rqet.find_tag_float(values_node, 'Value')
        elif values_type.startswith('Bool'):
            const_value = rqet.find_tag_bool(values_node, 'Value')
        else:
            const_value = rqet.find_tag_int(values_node, 'Value')

    return null_value, const_value



"""High level zonal grid function."""

import logging

log = logging.getLogger(__name__)

import os
import numpy as np

import resqpy.grid as grr
import resqpy.model as rq
import resqpy.olio.uuid as bu
import resqpy.olio.xml_et as rqet
import resqpy.property as rqp

from resqpy.derived_model._common import _write_grid, _establish_model_and_source_grid
from resqpy.derived_model._zone_layer_ranges_from_array import zone_layer_ranges_from_array


def zonal_grid(epc_file,
               source_grid = None,
               zone_title = None,
               zone_uuid = None,
               zone_layer_range_list = None,
               k0_min = None,
               k0_max = None,
               use_dominant_zone = False,
               inactive_laissez_faire = True,
               new_grid_title = None,
               new_epc_file = None):
    """Extends an existing model with a new version of the source grid converted to a single, thick, layer per zone.

    arguments:
       epc_file (string): file name to rewrite the model's xml to; if source grid is None, model is loaded from this file
       source_grid (grid.Grid object, optional): a multi-layer RESQML grid object; if None, the epc_file is loaded
          and it should contain one ijk grid object (or one 'ROOT' grid) which is used as the source grid
       zone_title (string): if not None, a discrete property with this as the citation title is used as the zone property
       zone_uuid (string or uuid): if not None, a discrete property with this uuid is used as the zone property (see notes)
       zone_layer_range_list (list of (int, int, int)): each entry being (min_k0, max_k0, zone_index); alternative to
          working from a zone array
       k0_min (int, optional): the minimum layer number in the source grid (zero based) to include in the zonal version;
          default is zero (ie. top layer in source grid)
       k0_max (int, optional): the maximum layer number in the source grid (zero based) to include in the zonal version;
          default is nk - 1 (ie. bottom layer in source grid)
       use_dominant_zone (boolean, default False): if True, the most common zone value in each layer is used for the whole
          layer; if False, then variation of zone values in active cells in a layer will raise an assertion error
       inactive_laissez_faire (boolean, optional): if True, a cell in the zonal grid will be set active if any of the
          corresponding cells in the source grid are active; otherwise all corresponding cells in the source grid
          must be active for the zonal cell to be active; default is True
       new_grid_title (string): used as the citation title text for the new grid object
       new_epc_file (string, optional): if None, the source epc_file is extended with the new grid object; if present,
          a new epc file (& associated h5 file) is created to contain the zonal grid (& crs)

    returns:
       new grid object (grid.Grid) with one layer per zone of the source grid

    notes:
       usually one of zone_title or zone_uuid or zone_layer_range_list should be passed, if none are passed then a
       single layer grid is generated; zone_layer_range_list will take precendence if present
    """

    assert epc_file or new_epc_file, 'epc file name not specified'
    if new_epc_file and epc_file and (
        (new_epc_file == epc_file) or
        (os.path.exists(new_epc_file) and os.path.exists(epc_file) and os.path.samefile(new_epc_file, epc_file))):
        new_epc_file = None
    model, source_grid = _establish_model_and_source_grid(epc_file, source_grid)
    assert source_grid.grid_representation in ['IjkGrid', 'IjkBlockGrid']
    if source_grid.grid_representation == 'IjkBlockGrid':
        source_grid.make_regular_points_cached()
    assert model is not None
    single_layer_mode = (not zone_title and not zone_uuid and
                         (zone_layer_range_list is None or len(zone_layer_range_list) == 1))

    k0_min, k0_max = _set_or_check_k_min_max(k0_min, k0_max, source_grid)

    if not single_layer_mode:  # process zone array
        if zone_layer_range_list is None:
            zone_array = _fetch_zone_array(source_grid, zone_title, zone_uuid)
            zone_layer_range_list = zone_layer_ranges_from_array(zone_array,
                                                                 k0_min,
                                                                 k0_max,
                                                                 use_dominant_zone = use_dominant_zone)
        zone_count = len(zone_layer_range_list)
        # above is list of (zone_min_k0, zone_max_k0, zone) sorted by zone_min_k0
        log.info('following layer ranges are based on top layer being numbered 1 (simulator protocol)')
        for (zone_min_k0, zone_max_k0, zone) in zone_layer_range_list:
            log.info('zone id {0:1d} covers layers {1:1d} to {2:1d}'.format(zone, zone_min_k0 + 1, zone_max_k0 + 1))
    else:
        zone_layer_range_list = [(k0_min, k0_max, 0)]
        zone_count = 1
    assert zone_count > 0, 'unexpected lack of zones'

    # create a new, empty grid object
    is_regular = grr.is_regular_grid(source_grid.root) and single_layer_mode
    grid = _empty_grid(model, source_grid, is_regular, k0_min, k0_max, zone_count)

    # aggregate inactive cell mask depending on laissez faire argument
    _set_inactive_cell_mask(source_grid, grid, inactive_laissez_faire, single_layer_mode, k0_min, k0_max,
                            zone_layer_range_list, zone_count)

    if not is_regular:
        _process_geometry(source_grid, grid, single_layer_mode, k0_min, k0_max, zone_layer_range_list, zone_count)

    # establish title for the new grid
    if new_grid_title is None or len(new_grid_title) == 0:
        if single_layer_mode:
            preamble = 'single layer'
        else:
            preamble = 'zonal'
        new_grid_title = preamble + ' version of ' + str(rqet.citation_title_for_node(source_grid.root))

    # write the new grid
    model.h5_release()
    if new_epc_file:
        _write_grid(new_epc_file, grid, grid_title = new_grid_title, mode = 'w')
    else:
        _write_grid(epc_file, grid, ext_uuid = None, grid_title = new_grid_title, mode = 'a')

    return grid


def single_layer_grid(epc_file,
                      source_grid = None,
                      k0_min = None,
                      k0_max = None,
                      inactive_laissez_faire = True,
                      new_grid_title = None,
                      new_epc_file = None):
    """Extends an existing model with a new version of the source grid converted to a single, thick, layer.

    arguments:
       epc_file (string): file name to rewrite the model's xml to; if source grid is None, model is loaded from this file
       source_grid (grid.Grid object, optional): a multi-layer RESQML grid object; if None, the epc_file is loaded
          and it should contain one ijk grid object (or one 'ROOT' grid) which is used as the source grid
       k0_min (int, optional): the minimum layer number in the source grid (zero based) to include in the single layer version;
          default is zero (ie. top layer in source grid)
       k0_max (int, optional): the maximum layer number in the source grid (zero based) to include in the single layer version;
          default is nk - 1 (ie. bottom layer in source grid)
       inactive_laissez_faire (boolean, optional): if True, a cell in the single layer grid will be set active if any of the
          corresponding cells in the source grid are active; otherwise all corresponding cells in the source grid
          must be active for the single layer cell to be active; default is True
       new_grid_title (string): used as the citation title text for the new grid object
       new_epc_file (string, optional): if None, the source epc_file is extended with the new grid object; if present,
          a new epc file (& associated h5 file) is created to contain the single layer grid (& crs)

    returns:
       new grid object (grid.Grid) with a single layer representation of the source grid
    """

    return zonal_grid(epc_file,
                      source_grid = source_grid,
                      k0_min = k0_min,
                      k0_max = k0_max,
                      inactive_laissez_faire = inactive_laissez_faire,
                      new_grid_title = new_grid_title,
                      new_epc_file = new_epc_file)


def _fetch_zone_array(grid, zone_title = None, zone_uuid = None, masked = True):
    properties = grid.extract_property_collection()
    assert properties is not None and properties.number_of_parts() > 0, 'no properties found in relation to grid'
    properties = rqp.selective_version_of_collection(properties, continuous = False)
    assert properties is not None and properties.number_of_parts() > 0,  \
       'no discreet properties found in relation to grid'
    if zone_title:
        properties = rqp.selective_version_of_collection(properties,
                                                         citation_title = zone_title)  # could make case insensitive?
        assert properties is not None and properties.number_of_parts() > 0,  \
           'no discreet property found with title ' + zone_title
    if zone_uuid:
        if isinstance(zone_uuid, str):
            zone_uuid = bu.uuid_from_string(zone_uuid)
        zone_uuid_str = str(zone_uuid)
        if zone_title:
            postamble = ' (and title ' + zone_title + ')'
        else:
            postamble = ''
        assert zone_uuid in properties.uuids(), 'no property found with uuid ' + zone_uuid_str + postamble
        part_name = grid.model.part(uuid = zone_uuid)
    else:
        part_name = properties.singleton()
    return properties.cached_part_array_ref(part_name, masked = masked)  # .copy() needed?


def _empty_grid(model, source_grid, is_regular, k0_min, k0_max, zone_count):
    if is_regular:
        dxyz_dkji = source_grid.block_dxyz_dkji.copy()
        dxyz_dkji[0] *= k0_max - k0_min + 1
        grid = grr.RegularGrid(model,
                               extent_kji = (1, source_grid.nj, source_grid.ni),
                               dxyz_dkji = dxyz_dkji,
                               origin = source_grid.block_origin,
                               crs_uuid = source_grid.crs_uuid,
                               set_points_cached = False)
    else:
        grid = grr.Grid(model)
        # inherit attributes from source grid
        grid.grid_representation = 'IjkGrid'
        grid.extent_kji = np.array((zone_count, source_grid.nj, source_grid.ni), dtype = 'int')
        grid.nk, grid.nj, grid.ni = zone_count, source_grid.nj, source_grid.ni
        grid.has_split_coordinate_lines = source_grid.has_split_coordinate_lines
        grid.crs_root = source_grid.crs_root
        grid.crs_uuid = source_grid.crs_uuid
    grid.k_direction_is_down = source_grid.k_direction_is_down
    grid.grid_is_right_handed = source_grid.grid_is_right_handed
    grid.pillar_shape = source_grid.pillar_shape
    return grid


def _scan_columns_for_reference_geometry(source_grid, grid, zone_layer_range_list, zone_count):
    source_points = source_grid.points_ref()
    log.debug('scanning columns (split pillars) for reference geometry')
    if not hasattr(source_grid, 'pillars_for_column'):
        source_grid.create_column_pillar_mapping()
    grid.pillars_for_column = source_grid.pillars_for_column.copy()
    for zone_i in range(zone_count):
        zk0_min, zk0_max = zone_layer_range_list[zone_i][0:2]
        for j in range(grid.nj):
            for i in range(grid.ni):
                if grid.inactive[zone_i, j, i]:
                    continue
                if zone_i == 0:
                    for k in range(zk0_min, zk0_max + 1):
                        if source_grid.array_cell_geometry_is_defined[k, j, i]:
                            for jp in range(2):
                                for ip in range(2):
                                    pillar = grid.pillars_for_column[j, i, jp, ip]
                                    grid.points_cached[0, pillar] = source_points[k, pillar]
                            break
                for k in range(zk0_max + 1, zk0_min - 1, -1):
                    if source_grid.array_cell_geometry_is_defined[k, j, i]:
                        for jp in range(2):
                            for ip in range(2):
                                pillar = grid.pillars_for_column[j, i, jp, ip]
                                grid.points_cached[zone_i + 1, pillar] = source_points[k + 1, pillar]
                        grid.array_cell_geometry_is_defined[zone_i, j, i] = True
                        break


def _set_inactive_cell_mask(source_grid, grid, inactive_laissez_faire, single_layer_mode, k0_min, k0_max,
                            zone_layer_range_list, zone_count):
    if source_grid.inactive is None:
        log.debug('setting inactive mask to None')
        grid.inactive = None
    elif single_layer_mode:
        if inactive_laissez_faire:
            log.debug('setting inactive mask using all mode (laissez faire)')
            grid.inactive = np.all(source_grid.inactive[k0_min:k0_max + 1], axis = 0).reshape(grid.extent_kji)
        else:
            log.debug('setting inactive mask using any mode (strict)')
            grid.inactive = np.any(source_grid.inactive[k0_min:k0_max + 1], axis = 0).reshape(grid.extent_kji)
    else:
        grid.inactive = np.zeros(grid.extent_kji, dtype = bool)
        for zone_i in range(zone_count):
            zk0_min, zk0_max, _ = zone_layer_range_list[zone_i]
            if inactive_laissez_faire:
                grid.inactive[zone_i] = np.all(source_grid.inactive[zk0_min:zk0_max + 1], axis = 0)
            else:
                grid.inactive[zone_i] = np.any(source_grid.inactive[zk0_min:zk0_max + 1], axis = 0)


def _process_geometry(source_grid, grid, single_layer_mode, k0_min, k0_max, zone_layer_range_list, zone_count):
    # rework the grid geometry
    source_grid.cache_all_geometry_arrays()
    # determine cell geometry is defined
    if hasattr(source_grid, 'array_cell_geometry_is_defined'):
        grid.array_cell_geometry_is_defined = np.empty(grid.extent_kji, dtype = bool)
        if single_layer_mode:
            grid.array_cell_geometry_is_defined[0] = np.logical_and(source_grid.array_cell_geometry_is_defined[k0_min],
                                                                    source_grid.array_cell_geometry_is_defined[k0_max])
        else:
            for zone_i in range(zone_count):
                zk0_min, zk0_max, _ = zone_layer_range_list[zone_i]
                grid.array_cell_geometry_is_defined[zone_i] = np.logical_and(
                    source_grid.array_cell_geometry_is_defined[zk0_min],
                    source_grid.array_cell_geometry_is_defined[zk0_max])
                # could attempt to pick up some corner points from K-neighbouring cells
        grid.geometry_defined_for_all_cells_cached = np.all(grid.array_cell_geometry_is_defined)
    else:
        grid.geometry_defined_for_all_cells_cached = source_grid.geometry_defined_for_all_cells_cached
    # copy info for pillar geometry is defined
    grid.geometry_defined_for_all_pillars_cached = source_grid.geometry_defined_for_all_pillars_cached
    if hasattr(source_grid, 'array_pillar_geometry_is_defined'):
        grid.array_pillar_geometry_is_defined = source_grid.array_pillar_geometry_is_defined.copy()
    # get reference to points for source grid geometry
    source_points = source_grid.points_ref()
    # slice top and base points
    points_shape = list(source_points.shape)
    points_shape[0] = zone_count + 1  # nk + 1
    grid.points_cached = np.zeros(points_shape)
    if grid.geometry_defined_for_all_cells_cached:
        if single_layer_mode:
            grid.points_cached[0] = source_points[k0_min]
            grid.points_cached[1] = source_points[k0_max + 1]  # base face
        else:
            for zone_i in range(zone_count):
                if zone_i == 0:
                    grid.points_cached[0] = source_points[zone_layer_range_list[zone_i][0]]
                grid.points_cached[zone_i + 1] = source_points[zone_layer_range_list[zone_i]
                                                               [1]]  # or could use 0th element of tuple for zone_i+1
    elif not grid.has_split_coordinate_lines:
        log.debug('scanning columns (unsplit pillars) for reference geometry')
        # fill in geometry: todo: replace with array operations if possible
        for zone_i in range(zone_count):
            zk0_min, zk0_max = zone_layer_range_list[zone_i][0:2]
            for j in range(grid.nj):
                for i in range(grid.ni):
                    if zone_i == 0:
                        for k in range(zk0_min, zk0_max + 1):
                            if source_grid.array_cell_geometry_is_defined[k, j, i]:
                                grid.points_cached[0, j:j + 2, i:i + 2] = source_points[k, j:j + 2, i:i + 2]
                                break
                    for k in range(zk0_max, zk0_min - 1, -1):
                        if source_grid.array_cell_geometry_is_defined[k, j, i]:
                            grid.points_cached[zone_i + 1, j:j + 2, i:i + 2] = source_points[k + 1, j:j + 2, i:i + 2]
                            grid.array_cell_geometry_is_defined[zone_i, j, i] = True
                            break
    else:
        _scan_columns_for_reference_geometry(source_grid, grid, zone_layer_range_list, zone_count)
    if grid.has_split_coordinate_lines:
        grid.split_pillar_indices_cached = source_grid.split_pillar_indices_cached.copy()
        grid.cols_for_split_pillars = source_grid.cols_for_split_pillars.copy()
        grid.cols_for_split_pillars_cl = source_grid.cols_for_split_pillars_cl.copy()
        grid.split_pillars_count = source_grid.split_pillars_count


def _set_or_check_k_min_max(k0_min, k0_max, source_grid):
    if k0_min is None:
        k0_min = 0
    else:
        assert k0_min >= 0 and k0_min < source_grid.nk
    if k0_max is None:
        k0_max = source_grid.nk - 1
    else:
        assert k0_max >= 0 and k0_max < source_grid.nk and k0_max >= k0_min
    return k0_min, k0_max
